# python
import warnings

# grAdapt
from .base import Initial


class Standard(Initial):
    """
    Simply uses a low discrepancy/equidistributed method and samples points by it.
    """
    def __init__(self, sampling_method=None):
        """
        Parameters
        ----------
        sampling_method : grAdapt.sampling.equidistributed Object
            Sample low discrepancy sequences when initial point method is not feasible
        """
        super().__init__(sampling_method)

    def sample(self, bounds, n_evals):
        """Returns a numpy array of sampled points.
        Does not include corner points of the hypercube/search space.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n_evals : int
            number of initial points sampled by method

        Returns
        -------
        (self.n_evals, len(self.bounds)) numpy array
        """
        super().sample(bounds, n_evals)
        return self.sampling_method.sample(bounds=bounds, n=n_evals)



import numpy as np


def random_starts(n_calls, dim):
    """Returns the number of random_starts which should be made
    Parameters
    ----------
    n_calls : number of iterations
    dim : number of dimensions of the function

    Returns
    -------
    numeric

    """
    # Exception handling
    if dim <= 0:
        raise Exception("Dimension dim must be equal or higher than 1.")

    if n_calls <= 0:
        raise Exception("n_evals must be equal or higher than 1.")

    fraction = 0.75 * (1 - (1 / np.log(n_calls + np.e))) * (1 - 1 / np.log(dim + np.e))
    return max(1, int(n_calls * fraction))


def is_inside_relative_range(value, ref_value, pct):
    """

    Parameters
    ----------
    value : numeric
    ref_value : numeric
    pct : numeric
        pct should be smaller than 1.

    Returns
    -------
    boolean
    """
    if ref_value * (1 - pct) <= value <= ref_value * (1 + pct):
        return True
    else:
        return False


def epochs(iteration):
    return min(100, int(iteration ** 1.25))



import time
import sys, os

# Disable
def blockPrint():
    sys.stdout = open(os.devnull, 'w')
    
# Restore
def enablePrint():
    sys.stdout = sys.__stdout__
    
# decorator to measure time
def timer(function):
    def new_function():
        start_time = time.time()
        blockPrint()
        params = function()
        elapsed = time.time() - start_time
        print('Time elapsed: {time}'.format(time=elapsed))
        enablePrint()
        return params
    return new_function


# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Datatype


class Float(Datatype):
    """Float Datatype
    Only the low and upper bound are required to define a Float Datatype.

    >>> from grAdapt.space.datatype import Float
    >>> first_dim = Float(low=10.5, high=20)
    >>> bounds = [first_dim]
    """

    def __init__(self, low, high, prior='uniform'):
        """

        Parameters
        ----------
        low: numeric
        high: numeric
        prior: str
            - if 'uniform' then nothing is changed
            - if 'log-uniform' then samples between low and high are drawn from the log-uniform distribution
        """

        # Exception handling
        if high < low:
            raise ValueError('high must be higher than low.')

        if prior == 'log-uniform':
            if low <= 0:
                raise ValueError('If logarithmic prior is used, then low must be higher than 0.')

        self.low = low
        self.high = high
        self.prior = prior
        self.bound = [low, high]
        self.len = 2
        self.dtype = 'float'

    def __len__(self):
        return self.len

    def __getitem__(self, key):
        return self.bound[key]

    def __setitem__(self, key, value):
        self.bound[key] = value

    def get_value(self, value):
        return self.get_float(value)

    def get_float(self, value):
        return value

    def __repr__(self):
        return 'Float('+str(self.bound)+')'

    def __str__(self):
        return 'Float('+str(self.bound)+')'

    def transform(self, x):
        """

        Parameters
        ----------
        x : numeric

        Returns
        -------
        numeric

        """
        if self.prior == 'log-uniform':
            # Inverse transform sampling using quantile function
            x_normalized = x / self.high
            return self.get_value(self.low * np.power(self.high * 1.0 / self.low, x_normalized))
        else:
            return self.get_value(x)



# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Equidistributed
from grAdapt.utils.sampling import sample_points_bounds
from grAdapt.utils.math.spatial import pairwise_distances


class Reassignment(Equidistributed):
    """Reassignment sampling method (slowest method)

    n Points are initially uniformly sampled given the bounds. The point
    with the smallest distance to another is being resampled uniformly.
    Avoids 'evil neighbours' but is slowest.
    For a better performance: Sliding window has been added.

    """

    def __init__(self, n_iters=None, window_size=500):
        """
        Parameters
        ----------
        n_iters : integer
            number of iterations to reassign worse points
        window_size : integer
            size of history points to consider. smaller is faster but worsen the results.
        """
        super().__init__()
        self.n_iters = n_iters
        self.window_size = window_size

    def sample(self, bounds, n, x_history=None):
        """Samples low discrepancy/equidistributed sequences
        Method has to handle with new bounds and n.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n : int
            number of points to be sampled
        x_history : array-like (2d)
            History points. Consider those to prevent sampling in dense regions.

        Returns
        -------
        array-like (n, len(bounds))
            Returns a 2D array. dim is the dimension of a single point
            Each row corresponds to a single point.
            Each column corresponds to a dimension.
        """

        # set to new variables
        super().sample(bounds, n, x_history)

        if self.n_iters is None:
            self.n_iters = min(int(self.n//2), 1)

        if x_history is None:
            x_history = sample_points_bounds(self.bounds, 1).reshape(1, -1)

        p = sample_points_bounds(self.bounds, self.n)
        p_x_history = np.vstack((x_history, p))
        for i in range(self.n_iters):
            dists_matrix = pairwise_distances(p_x_history[-self.window_size:], p_x_history[-self.window_size:])
            np.fill_diagonal(dists_matrix, np.inf)
            idx_min_eps = np.argmin(np.min(dists_matrix, axis=1))
            p_x_history[-self.window_size:][idx_min_eps] = sample_points_bounds(self.bounds, 1)

        return p_x_history[-self.n:]



# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Equidistributed
from grAdapt.utils.sampling import sample_points_bounds
from grAdapt.utils.math.spatial import pairwise_distances


class MaximalMinDistance(Equidistributed):
    """Maximal min distance sampling method

    A fixed amount of points are candidates. The candidate is chosen
    as a point if it has the highest minimal epsilon margin among
    other points. The minimal epsilon margin is the smallest distance
    between two points. Each point has an minimal epsilon margin.
    For a better performance, only a fixed amount of latest points
    are considered.
    Has a disadvantage: creates 'evil' neighbours.

    """

    def __init__(self, n_candidates=10, window_size=500):
        """
        Parameters
        ----------
        n_candidates : integer
            number of candidates
        window_size : integer
            size of history points to consider. smaller is faster but worsen the results.
        """
        super().__init__()
        self.n_candidates = n_candidates
        self.window_size = window_size

    def sample(self, bounds, n, x_history=None):
        """Samples low discrepancy/equidistributed sequences
        Method has to handle with new bounds and n.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n : int
            number of points to be sampled
        x_history : array-like (2d)
            History points. Consider those to prevent sampling in dense regions.

        Returns
        -------
        array-like (n, len(bounds))
            Returns a 2D array. dim is the dimension of a single point
            Each row corresponds to a single point.
            Each column corresponds to a dimension.
        """

        # set to new variables
        super().sample(bounds, n, x_history)

        if x_history is None:
            x_history = sample_points_bounds(self.bounds, 1)
            x_history_list = list(x_history)
        else:
            x_history_list = list(x_history)

        for i in range(self.n):
            x_history_sublist = x_history_list[-self.window_size:]
            candidates = sample_points_bounds(self.bounds, self.n_candidates)
            dists_matrix = pairwise_distances(candidates, np.array(x_history_sublist))
            min_dists = np.min(dists_matrix, axis=1)
            max_min_dists = np.argmax(min_dists)
            x_history_list.append(candidates[max_min_dists])

        return np.array(x_history_list)[-self.n:]



# decorators
from abc import abstractmethod


class Initial:
    """
    Initial Base class
    """

    def __init__(self, sampling_method=None):
        """
        Parameters
        ----------
        sampling_method : grAdapt.sampling.equidistributed Object
            Sample low discrepancy sequences when initial point method is not feasible

        """
        self.bounds = None
        self.n_evals = None
        self.sampling_method = sampling_method

    @abstractmethod
    def sample(self, bounds, n_evals):
        """Returns a numpy array of sampled points.
        Might include corners of the hypercube depending on the method.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n_evals : int
            number of initial points sampled by method

        Returns
        -------
        (self.n_evals, len(self.bounds)) numpy array
        """
        self.bounds = bounds
        self.n_evals = n_evals



# Python Standard Libraries
# decorators
from abc import abstractmethod
from deprecated import deprecated
import warnings

# ignore scipy warnings for GPR
warnings.filterwarnings('ignore')

# Third party imports
import numpy as np
import scipy.linalg
from sklearn.gaussian_process import GaussianProcessRegressor

# Package imports
from .base import Surrogate
from .kernels import Nystroem, RationalQuadratic  # , RBF

# grAdapt
from grAdapt.utils.math.linalg import inv_stable


class GPR(Surrogate):
    """Gaussian Process Regression surrogate
    Implements various gradient calculations depending
    on the chosen kernel.
    """

    def __call__(self):
        pass

    def __init__(self, **surrogate_params):
        """
        Parameters
        ----------
        surrogate_params : dictionary
        """
        if surrogate_params is None or len(surrogate_params) == 0:
            surrogate_params = {'kernel': RationalQuadratic()}
            # surrogate_params = {'kernel': Nystroem(RationalQuadratic)}
        super().__init__(GaussianProcessRegressor(**surrogate_params), surrogate_params)
        self.kernel_function = surrogate_params['kernel']
        self.alpha = None

    @abstractmethod
    def fit(self, X, y):
        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        kernel_matrix = self.kernel_function(X, X)
        kernel_matrix_inv = inv_stable(kernel_matrix)
        self.alpha = kernel_matrix_inv @ y

    def fit_scikit(self, X, y):
        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        self.model.fit(X, y)
        self.alpha = self.model.alpha_

    @abstractmethod
    def predict(self, x_predict, **args):
        if len(x_predict.shape) == 1:
            x_predict = x_predict.reshape(1, -1)
        return self.model.predict(x_predict, **args)

    @deprecated("This function is deprecated. The gradient evaluation is not vectorized.")
    def eval_gradient_old(self, x, surrogate_grad_params):
        """Evaluates the gradient of GPR at point x
        Parameters
        ----------
        x: 1D array-like (d,)
        surrogate_grad_params: list
            surrogate_grad_params[0]: whole training data shape (m, d)


        Returns
        -------
        grad : 1D array-like
            same shape as x
        """

        x_train = surrogate_grad_params[0]
        x = x.reshape(1, -1)
        kernel_matrix = self.kernel_function(x, x_train)
        length_scale = self.kernel_function.length_scale
        n, m = kernel_matrix.shape[0], kernel_matrix.shape[1]
        d = x.shape[1]
        grad = np.zeros(d)

        # calculate gradient
        for k in range(d):
            tmp_kernel = np.copy(kernel_matrix)
            for j in range(m):
                tmp_kernel[0, j] = ((x_train[j, k] - x[0, k]) * tmp_kernel[0, j]) / (length_scale ** 2)

            grad[k] = tmp_kernel.dot(self.alpha)

        return grad

    def grad_RBF(self, x, surrogate_grad_params):
        x_train = surrogate_grad_params[0]
        x = x.reshape(1, -1)
        kernel_matrix = self.kernel_function(x, x_train)
        length_scale = self.kernel_function.length_scale
        n, m = kernel_matrix.shape[0], kernel_matrix.shape[1]
        d = x.shape[1]
        grad = np.zeros(d)

        # calculate gradient
        for k in range(d):
            tmp_kernel = np.copy(kernel_matrix)
            tmp_kernel[0, :] = ((x_train[:, k] - x[0, k]) * tmp_kernel[0, :]) / (length_scale ** 2)

            grad[k] = tmp_kernel.dot(self.alpha)

        return grad

    def grad_RationalQuadratic(self, x, surrogate_grad_params):
        x_train = surrogate_grad_params[0]
        x = x.reshape(1, -1)
        kernel_matrix = self.kernel_function(x, x_train)
        length_scale = self.kernel_function.length_scale
        alpha = self.kernel_function.alpha  # this alpha here is a parameter for the kernel!
        n, m = kernel_matrix.shape[0], kernel_matrix.shape[1]
        d = x.shape[1]
        grad = np.zeros(d)

        # calculate gradient
        for k in range(d):
            tmp_kernel = np.copy(kernel_matrix)
            d2 = (np.sum((x[0, k] - x_train[:, k]) ** 2)) / (2 * alpha * length_scale ** 2)
            tmp_kernel[0, :] = -(2 * x[0, k] - 2 * x_train[:, k]) / (2 * length_scale ** 2) * (
                    tmp_kernel[0, :] / (1 + d2))

            grad[k] = tmp_kernel.dot(self.alpha)

        return grad

    def eval_gradient(self, x, surrogate_grad_params):
        """Evaluates the gradient of GPR at point x
                Parameters
                ----------
                x: 1D array-like (d,)
                surrogate_grad_params: list
                    surrogate_grad_params[0]: whole training data shape (m, d)


                Returns
                -------
                grad : 1D array-like
                    same shape as x
        """

        kernel_name = self.kernel_function.__str__()
        if 'Nystroem' in self.kernel_function.__str__():
            kernel_name = self.kernel_function.kernel_name

        if 'RationalQuadratic' in kernel_name:
            return self.grad_RationalQuadratic(x, surrogate_grad_params)
        elif any(list(map(lambda x: x in kernel_name, ['None', 'RBF']))):
            return self.grad_RBF(x, surrogate_grad_params)
        else:
            raise AttributeError('Given kernel function of surrogate model (GPR) is not supported. '
                                 'Set a different Kernel. '
                                 'Used Kernel: ' + kernel_name())


class GPROnlineLearningWrapper(GPR):
    """Gaussian Process Regression (Subclass of GPR)
    Do not use this class!
    Superclass of many Online Learning Implementations
    for GPR.

    """

    def __init__(self, **surrogate_params):
        """
        Parameters
        ----------
        surrogate_params : dictionary
        """
        if surrogate_params is None or len(surrogate_params) == 0:
            surrogate_params = {'kernel': RationalQuadratic()}
            #surrogate_params = {'kernel': Nystroem(RationalQuadratic)}
        self.surrogate_params = surrogate_params
        self.kernel_function = surrogate_params['kernel']
        self.kernel_matrix = None
        self.kernel_matrix_inv = None
        self.alpha = None
        self.X_train = None

    def insert(self, X1, X2=None):
        """Calculates Covariance/Kernel matrix and the inverse of it
        Instead of recalculating, Kernel Matrix is updated
        by stacking a row and a column on the right

        Parameters
        ----------
        X1 : array-like 2D
        X2 : array-like 2D

        Returns
        -------

        """
        if X2 is None:
            X2 = X1

        # Training continuation
        if self.kernel_matrix is None:
            self.kernel_matrix = self.kernel_function(X1, X2)
            self.kernel_matrix_inv = inv_stable(self.kernel_matrix)
        else:
            b = self.kernel_function(X1[-1].reshape(1, -1), X2)
            self.kernel_matrix = np.concatenate((self.kernel_matrix, b[0, :-1].reshape(1, -1)), axis=0)
            self.kernel_matrix = np.concatenate((self.kernel_matrix, b.T), axis=1)

            # Inversion of kernel matrix
            # Augmented Inverse Update
            d = b[0, -1].reshape(1, 1)
            b = b[0, :-1].reshape(-1, 1)
            e = self.kernel_matrix_inv @ b
            g = 1 / (d - b.T @ e)

            self.kernel_matrix_inv = self.kernel_matrix_inv + g * e @ e.T
            row_add = (-g * e).T
            self.kernel_matrix_inv = np.concatenate((self.kernel_matrix_inv, row_add.reshape(1, -1)),
                                                    axis=0)
            column_add = np.concatenate((row_add, g), axis=1)
            self.kernel_matrix_inv = np.concatenate((self.kernel_matrix_inv, column_add.T), axis=1)

    def delete(self, X1, X2=None):
        """Calculates Covariance/Kernel matrix and the inverse of it
        Instead of recalculating, Kernel Matrix is updated

        Parameters
        ----------
        X1 : array-like 2D
        X2 : array-like 2D

        Returns
        -------

        """
        if X2 is None:
            X2 = X1

        # Training continuation
        if self.kernel_matrix is None:
            self.kernel_matrix = self.kernel_function(X1, X2)
            self.kernel_matrix_inv = inv_stable(self.kernel_matrix)
        else:
            # delete first row and column
            self.kernel_matrix = self.kernel_matrix[1:, 1:]

            # insert new data point
            h = self.kernel_function(X1[-1].reshape(1, -1), X2[-self.window_size:, :]).T
            g = h[-1, 0]
            self.kernel_matrix = np.concatenate((self.kernel_matrix, h[:-1, 0].reshape(-1, 1).T), axis=0)
            self.kernel_matrix = np.concatenate((self.kernel_matrix, h), axis=1)

            # Inversion of kernel matrix
            # Augmented Inverse Update
            e = self.kernel_matrix_inv[0, 0]
            f = self.kernel_matrix_inv[1:, 0]
            H = self.kernel_matrix_inv[1:, 1:]

            B = H - (f @ f.T) / e
            h = h[:-1, :]
            s = 1 / (g - h.T @ B @ h)
            self.kernel_matrix_inv = B + (B @ h) @ (B @ h).T * s

            # add rows and columns
            row_add = - (B @ h).T * s
            column_add = np.concatenate((row_add, s), axis=1)
            self.kernel_matrix_inv = np.concatenate((self.kernel_matrix_inv, row_add), axis=0)
            self.kernel_matrix_inv = np.concatenate((self.kernel_matrix_inv, column_add.T), axis=1)

    @abstractmethod
    def fit(self, X, y):
        pass

    def predict(self, x_predict, return_std=False):
        if self.kernel_matrix is None:
            raise Exception('Make sure to fit model before using predict.')

        if len(x_predict.shape) == 1:
            x_predict = x_predict.reshape(1, -1)

        # positive definite matrix
        Id = np.eye(self.kernel_matrix.shape[0])
        self.kernel_matrix = self.kernel_matrix + Id * 1e-7

        pred_matrix = self.kernel_function(x_predict, self.X_train)

        if return_std:
            sigma_22 = self.kernel_function(x_predict)
            solved = scipy.linalg.solve(self.kernel_matrix, pred_matrix.T, assume_a='pos').T
            posterior_cov = sigma_22 - solved @ pred_matrix.T
            sigma = np.sqrt(np.maximum(0, np.diag(posterior_cov)))
            return pred_matrix @ self.alpha, sigma
        else:
            return pred_matrix @ self.alpha

    def negative_marginal_likelihood(self):
        pass


class GPROnlineInsert(GPROnlineLearningWrapper):
    """Gaussian Process Regression (Subclass of GPROnlineLearningWrapper)
    GPR Online Learning with Augmented Inverse Update
    @ Formulas derived by https://github.com/Bigpig4396
    @ Original paper: Incremental Sparsification for Real-time Online Model Learning (Nguyen, Peters)

    In order to fit a new point, a history of points is required i.e.:
    >>> import numpy as np
    >>> # Initialize Training Data
    >>> x = np.linspace(-10, 10).reshape(-1, 1)
    >>> y = np.random.random(x.shape[0])
    >>> # Fit with GPROnlineInsert
    >>> gpr = GPROnlineInsert()
    >>> for i in range(x.shape[0]):
    >>>     gpr.fit(x[:i], y[:i])
    """

    def __init__(self, **surrogate_params):
        """
        Parameters
        ----------
        surrogate_params : dictionary
        """
        super().__init__(**surrogate_params)

    def fit(self, X, y):
        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        self.X_train = X

        if self.kernel_matrix is not None:
            for i in range(self.kernel_matrix.shape[0], self.X_train.shape[0]):
                self.insert(X[:i + 1, :])
        else:
            self.insert(X)
        self.alpha = self.kernel_matrix_inv @ y


class GPRSlidingWindow(GPROnlineLearningWrapper):
    """Gaussian Process Regression (Subclass of GPROnlineLearningWrapper)
    GPR Online Learning with Sliding Window

    @ Formulas derived by https://github.com/Bigpig4396
    @ Original papers :
                        Incremental Sparsification for Real-time Online Model Learning (Nguyen, Peters, 2011)
                        Kernel Adaptive Learning (Principe, Liu, 2008)

    Only keeps the latest window_size points where window_size is a user-defined parameter.

    In order to fit a new point, a history of points is required i.e.:
    >>> import numpy as np
    >>> # Initialize Training Data
    >>> x = np.linspace(-10, 10, 1000).reshape(-1, 1)
    >>> y = np.random.random(x.shape[0])
    >>> # Fit with GPROnlineInsert
    >>> gpr = GPRSlidingWindow(window_size=100)
    >>> for i in range(x.shape[0]):
    >>>     gpr.fit(x[:i], y[:i])
    """

    def __init__(self, **surrogate_params):
        """
        Parameters
        ----------
        surrogate_params : dictionary
        """
        super().__init__(**surrogate_params)
        if 'window_size' not in self.surrogate_params:
            self.surrogate_params['window_size'] = 300
        self.window_size = self.surrogate_params['window_size']

    def fit(self, X, y):
        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        self.X_train = X

        if self.X_train.shape[0] > self.window_size:
            X = X[-self.window_size:, :]
            self.delete(X, X)
        else:
            # insert with training continuation
            if self.kernel_matrix is not None:
                for i in range(self.kernel_matrix.shape[0], self.X_train.shape[0]):
                    self.insert(X[:i + 1, :], X[:i + 1, :])
            else:
                self.insert(X, X)

        self.alpha = self.kernel_matrix_inv @ y[-self.window_size:]

    def predict(self, x_predict, return_std=False):
        self.X_train = self.X_train[-self.window_size:, :]
        return super().predict(x_predict, return_std)

    def grad_RBF(self, x, surrogate_grad_params):
        surrogate_grad_params_window = surrogate_grad_params.copy()
        x_train = surrogate_grad_params_window[0]
        if x_train.shape[0] > self.window_size:
            x_train = x_train[-self.window_size:, :]

        surrogate_grad_params_window[0] = x_train
        return super().grad_RBF(x, surrogate_grad_params_window)

    def grad_RationalQuadratic(self, x, surrogate_grad_params):
        surrogate_grad_params_window = surrogate_grad_params.copy()
        x_train = surrogate_grad_params_window[0]
        if x_train.shape[0] > self.window_size:
            x_train = x_train[-self.window_size:, :]

        surrogate_grad_params_window[0] = x_train
        return super().grad_RationalQuadratic(x, surrogate_grad_params_window)


class GPROnlineStandard(GPROnlineLearningWrapper):
    """Gaussian Process Regression surrogate
    GPR Online Learning with Augmented Inverse Update
    @ Formulas derived by https://github.com/Bigpig4396
    @ Original paper: Incremental Sparsification for Real-time Online Model Learning (Nguyen, Peters)

    This class is not meant to be used with Sequential. Use GPROnlineInsert instead.
    It is meant to be used outside of grAdapt.

    Fit single points, no history of data points needed i.e.:
    >>> gpr = GPROnlineStandard()
    >>> gpr.fit(x[:x.shape[0]-1], y[:x.shape[0]-1])
    >>> gpr.fit_online(x[x.shape[0]], y[x.shape[0]])
    """

    def __init__(self, **surrogate_params):
        """
        Parameters
        ----------
        surrogate_params : dictionary
        """
        super().__init__(**surrogate_params)
        self.y_train = None

    def kernel(self, X1, X2=None):
        """Calculates Covariance/Kernel matrix and the inverse of it
        Instead of recalculating, Kernel Matrix is updated
        by stacking a row and a column on the right
        Does not need a history of training data!

        Parameters
        ----------
        X1 : array-like 2D
        X2 : array-like 2D

        Returns
        -------

        """
        if X2 is None:
            X2 = self.X_train

        b = self.kernel_function(X1, X2)
        self.kernel_matrix = np.concatenate((self.kernel_matrix, b[0, :-1].reshape(1, -1)), axis=0)
        self.kernel_matrix = np.concatenate((self.kernel_matrix, b.T), axis=1)

        # Inversion of kernel matrix
        # Augmented Inverse Update
        d = b[0, -1].reshape(1, 1)
        b = b[0, :-1].reshape(-1, 1)
        e = self.kernel_matrix_inv @ b
        g = 1 / (d - b.T @ e)

        self.kernel_matrix_inv = self.kernel_matrix_inv + g * e @ e.T
        row_add = (-g * e).T
        self.kernel_matrix_inv = np.concatenate((self.kernel_matrix_inv, row_add.reshape(1, -1)),
                                                axis=0)
        column_add = np.concatenate((row_add, g), axis=1)
        self.kernel_matrix_inv = np.concatenate((self.kernel_matrix_inv, column_add.T), axis=1)

    def fit(self, X, y):
        """First fit or refit points
        This method does not do online/incremental learning!

        Parameters
        ----------
        X : array-like 2D (n, d)
        y : array-like 1d (n, )

        Returns
        -------

        """
        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        self.X_train = X
        self.y_train = y
        self.kernel_matrix = self.kernel_function(X, X)
        self.kernel_matrix_inv = inv_stable(self.kernel_matrix)
        self.alpha = self.kernel_matrix_inv @ self.y_train

    # TODO: Bug here?
    def fit_online(self, X, y):
        """Fits to GPR online/incremental
        Can only fit a single point.

        Parameters
        ----------
        X : array-like 2D (n, d)
        y : array-like 1d (n, )

        Returns
        -------

        """
        if len(X.shape) == 1:
            X = X.reshape(1, -1)
        
        self.y_train = np.concatenate((self.y_train, y), axis=0)

        if self.X_train is None:
            self.fit(X, y)
        else:
            initial_size = self.X_train.shape[0]
            self.X_train = np.concatenate((self.X_train, X), axis=0)
            for i in range(X.shape[0]):
                self.kernel(X[i].reshape(1, -1), self.X_train[: i + initial_size + 1])

            self.alpha = self.kernel_matrix_inv @ self.y_train



from __future__ import absolute_import

from . import datasets
from . import escape
from . import models
from . import optimizer
from . import sampling
from . import space
from . import surrogate
from . import utils

# attributes
__version__ = '0.1.1a1'



from .base import Equidistributed

from .Density import Density
from .KDistance import KDistance
from .KDistanceMean import KDistanceMean
from .MaximalMinDistance import MaximalMinDistance
from .Mitchell import Mitchell
from .Reassignment import Reassignment



from .base import Optimizer

from .AdaGradBisection import AdaGrad, AdaGradBisection
from .AdaMaxBisection import AdaMax, AdaMaxBisection
from .AdamBisection import Adam, AdamBisection
from .AMSGradBisection import AMSGrad, AMSGradBisection
from .GradientDescentBisection import GradientDescent, GradientDescentBisection
from .MomentumBisection import Momentum, MomentumBisection
from .RMSPropBisection import RMSProp, RMSPropBisection



# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Datatype


class Nominal(Datatype):
    """Nominal Datatype
    Is used for nominal categories. The categories are not ordered. Example usage:

    >>> from grAdapt.space.datatype import Nominal
    >>> animals = Nominal(['cat', 'dog', '´turtle'])
    >>> bounds = [animals]

    Internally, each category is a dimension of its own. The categories are encoded to one-hot.
    """

    def __init__(self, list_str):
        """

        Parameters
        ----------
        list_str : list of strings
            Each string represents a category
        """
        raise NotImplementedError
        self.len = len(list_str)
        self.bound = [0, self.len-1]
        self.categories = list_str
        self.dtype = 'nominal'

    def __len__(self):
        return self.len

    def __getitem__(self, key):
        return self.bound[key]

    def __setitem__(self, key, value):
        self.bound[key] = value

    def __repr__(self):
        return 'Nominal('+str(self.categories)+')'

    def __str__(self):
        return 'Nominal'+str(self.categories)+')'

    def get_value(self, value):
        # value is onehot encoded
        return self.get_category(value)

    def get_category(self, value):
        # value is onehot encoded
        key = np.argmax(value)
        return self.categories[key]

    def transform(self, x):
        """

        Parameters
        ----------
        x : numeric

        Returns
        -------
        numeric

        """
        closest_category = np.floor(x)
        onehot_encoding = np.zeros((len(self.len),))
        onehot_encoding[closest_category] = 1
        return onehot_encoding



# Python Standard Libraries
# decorators
import warnings
from deprecated import deprecated

# Third party imports
import numpy as np

# Package imports
from .base import Surrogate


class NoModel(Surrogate):
    """NoModel Surrogate
    No surrogate model is used here
    The gradients are approximated by partial difference quotient
    @ https://en.wikipedia.org/wiki/Partial_derivative
    Number of function evaluations is n_evals * dim * 100

    h : Parameter based on search space
    """
    def __init__(self, **surrogate_params):
        super().__init__(None, surrogate_params)
        self.X_train = None
        warnings.warn('You are using no surrogate model. The number of function'
                      ' evaluations is therefore much higher than expected. However, it'
                      ' can perform better on higher dimensions.')

    def fit(self, X, y):
        """Fit training data (X, y)
        Parameters
        ----------
        X : array-like (n, d)
        y : array-like (n,)


        Returns
        -------
        None

        """

        if len(X.shape) == 1:
            raise Exception('If \'NoModel\' is used, please enter a higher function'
                            ' evaluation number \'n_evals\'.')
        else:
            self.X_train = X

    def predict(self, x_predict, return_std=False):
        """Predict y value based on x_predict
        Parameters
        ----------
        x_predict : array-like (n, d)
        return_std : array-like (n,)
            returns the standard deviation of the posterior mean


        Returns
        -------
        array-like (n,)

        """
        if return_std:
            return np.zeros_like(x_predict), np.zeros_like(x_predict)
        else:
            return np.zeros_like(x_predict)

    def eval_gradient(self, x, surrogate_grad_params):
        """Evaluates the gradient of GPR at point x
        Parameters
        ----------
        x: 1D array-like (d,)
        surrogate_grad_params: list
            surrogate_grad_params[0]: whole training data shape (m, d)

        Returns
        -------
        grad : 1D array-like
            same shape as x
        """
        # x_train = surrogate_grad_params[0]
        # y_train = surrogate_grad_params[1]
        func = surrogate_grad_params[2]
        bounds = surrogate_grad_params[3]

        grad = np.zeros_like(x)
        x_copy = np.copy(x)
        # difference quotient
        for i in range(x.shape[0]):
            hi = np.min(np.array([np.abs(bounds[i][0]-bounds[i][1])/100, 1e-2]))
            original_value = x[i]
            x_tilde = x

            # perturbation of xi with h
            xi_perturbation = x_tilde[i] + hi
            if not(bounds[i][0] <= xi_perturbation <= bounds[i][1]):
                xi_perturbation = x_tilde[i] - hi
                hi = -hi
            x_tilde[i] = xi_perturbation

            # gradient
            if hi != 0:
                dfdxi = (func(x_tilde)-func(x_copy))/hi
            else:
                dfdxi = 0

            grad[i] = dfdxi

            # revert back to original value
            x_tilde[i] = original_value
            x[i] = original_value

        return grad

    @deprecated("This function is deprecated and very slow due to np.copy")
    def eval_gradient_deprecated(self, x, surrogate_grad_params):
        """Evaluates the gradient of GPR at point x
        Parameters
        ----------
        x: 1D array-like (d,)
        surrogate_grad_params: list
            surrogate_grad_params[0]: whole training data shape (m, d)


        Returns
        -------
        grad : 1D array-like
            same shape as x
        """
        # x_train = surrogate_grad_params[0]
        # y_train = surrogate_grad_params[1]
        func = surrogate_grad_params[2]
        bounds = surrogate_grad_params[3]

        grad = np.zeros_like(x)
        # difference quotient
        for i in range(x.shape[0]):
            hi = np.min(np.array([np.abs(bounds[i][0]-bounds[i][1])/100, 1e-2]))
            x_tilde = np.copy(x)

            # perturbation of xi with h
            xi_perturbation = x_tilde[i] + hi
            if not(bounds[i][0] <= xi_perturbation <= bounds[i][1]):
                xi_perturbation = x_tilde[i] - hi
                hi = -hi
            x_tilde[i] = xi_perturbation

            # gradient
            if hi != 0:
                dfdxi = (func(x_tilde)-func(x))/hi
            else:
                dfdxi = 0

            grad[i] = dfdxi

        return grad



# Python Standard Libraries
from abc import abstractmethod

# Third party imports
import numpy as np

from .base import Optimizer


class RMSProp(Optimizer):
    """Optimizer object

    Implementation of RMSProp

    """
    def __init__(self, surrogate, params=[1e-3, 0.9]):
        super().__init__(surrogate, params)

    def run(self, xp, num_iters, surrogate_grad_params):
        alpha = self.params[0] # learning rate
        beta = self.params[1]
        eps = 1e-3 # catch dividing by zero
        x_next = xp
        mean_square = np.zeros_like(xp)

        for i in range(num_iters):
            grad = self.surrogate.eval_gradient(x_next, surrogate_grad_params)
            mean_square = beta * mean_square + (1 - beta) * grad**2
            x_old = x_next
            x_next = x_next - alpha * grad / (np.sqrt(mean_square) + eps)

            if np.linalg.norm(x_next - x_old) < 1e-3:
                return x_next
        return x_next


class RMSPropBisection(Optimizer):
    """RMSProp with Bisection
    Do not use! It does not work well with Bisection method!
    """
    def __init__(self, surrogate, params=[1e-5, 0.9]):
        super().__init__(surrogate, params)

    def run(self, xp, num_iters, surrogate_grad_params, k=1):
        alpha = self.params[0]  # learning rate
        beta = self.params[1]
        eps = 1e-3  # catch dividing by zero
        x_next = xp
        mean_square = np.zeros_like(xp)
        grad = self.surrogate.eval_gradient(x_next, surrogate_grad_params)

        for i in range(1, num_iters):
            # store old point with its gradient
            x_old = x_next
            grad_old = grad

            # RMSProp update rule
            mean_square = beta * mean_square + (1 - beta) * grad_old ** 2
            x_new = x_next - alpha * grad_old / (np.sqrt(mean_square) + eps)
            grad = self.surrogate.eval_gradient(x_next, surrogate_grad_params)

            # Bisection
            # sign changed happened if sign is negative
            sign = grad * grad_old
            sign_changed = sign < 0
            # adapt alpha learning rate
            alpha = sign_changed * alpha / (k**2) + (1 - sign_changed) * alpha * k
            # x_next lies between x_old and x_new
            x_next = (1 - sign_changed) * x_new + sign_changed * (x_new + x_old) / 2
            #print(x_next)
            if np.linalg.norm(x_next - x_old) < 1e-3:
                return x_next
        return x_next



import numpy as np


def pairwise_distances(p, q):
    first_term = np.einsum('ij->i', p ** 2)  # sum along dimension
    second_term = np.einsum('ij->i', q ** 2)  # sum along dimension
    third_term = np.einsum('ik,jk->ij', p, q)  # pairwise product
    squared_distances = first_term[:, None] + second_term
    squared_distances -= 2 * third_term

    # Ignore numerical error
    squared_distances[squared_distances < 1e-10] = 0

    return np.sqrt(squared_distances)


# Python Standard Libraries
from itertools import product
import warnings

# Third party imports
import numpy as np

from grAdapt.utils.math.spatial import pairwise_distances


def sample_corner_bounds(bounds):
    """Samples points on the corner of given bounds

    Parameters
    ----------
    bounds : list of tuples.
        Each tuple in the list defines the bounds for the corresponding variable
        Example: [(1, 2), (2, 3), (-1, 4)...]

    Returns
    -------
    array (2**len(bounds), len(bounds))

    Examples
    --------
    >>> import grAdapt.utils.sampling as sampling

    Simple example:

    >>> bounds = [(-1, 1), (2, 4)]
    >>> corner_points = sampling.sample_corner_bounds(bounds)
    >>> print(corner_points)
    array([[-1.  2.]
           [-1.  4.]
           [ 1.  2.]
           [ 1.  4.]])

    Using grAdapt Datetype objects:

    >>> from grAdapt.space.datatype import Float, Integer
    >>> range1 = Float(-1.4, 2.3)
    >>> range2 = Integer(2, 3)
    >>> bounds = [range1, range2]
    >>> corner_points = sampling.sample_corner_bounds(bounds)
    >>> print(corner_points)
    array([[-1.4  2. ]
           [-1.4  3. ]
           [ 2.3  2. ]
           [ 2.3  3. ]])
    """
    warnings.warn('Sampling corner points can cause memory leaks. Number of corner points rises exponentially with '
                  'each additional dimension. Use with caution.', ResourceWarning)

    return np.array(list(product(*bounds)), dtype=np.float)


def sample_points_bounds(bounds, n=1, random_state=None):
    """Random sample points uniformly given bounds

    Parameters
    ----------
    bounds : list of tuples.
        Each tuple in the list defines the bounds for the corresponding variable
        Example: [(1, 2), (2, 3), (-1, 4)...]
    n : integer
        Describes the number of points which should be sampled from the given
        bound.
    random_state : int
        random seed

    Returns
    -------
    sampled_points : array-like (n, dim)
        Returns a 2D array. dim is the dimension of a single point
        Each row corresponds to a single point.
        Each column corresponds to a dimension.
    """
    # setting the seed
    if random_state is not None:
        np.random.seed(random_state)

    if n <= 0:
        raise Exception("n should be higher than 0.")
    
    dim = len(bounds)
    sampled_points = np.empty((n, dim))
    for j in range(dim):
        sampled_points[:, j] = np.random.uniform(bounds[j][0], bounds[j][1], (n,))

    return sampled_points


# @jit(nopython=True)
def inside_bounds(bounds, x_next):
    """Checks whether the point x_next is inside of bounds

    Parameters
    ----------
    bounds : list of tuples
        Each tuple in the list defines the bounds for the corresponding variable
        Example: [(1, 2), (2, 3), (-1, 4)...]
    x_next : array_like (d, )


    Returns
    -------
    boolean

    """
    n = len(bounds)
    for i in range(n):
        if not (bounds[i][0] <= x_next[i] <= bounds[i][1]):
            return False

    return True


def inside_bounds_ndim(bounds, x):
    """Checks whether all n samples in x lie inside of bounds

    Parameters
    ----------
    bounds : list of tuples
        Each tuple in the list defines the bounds for the corresponding variable
        Example: [(1, 2), (2, 3), (-1, 4)...]
    x : array_like (n, d)
        n samples with d dimensions


    Returns
    -------
    boolean

    """
    return all(list(map(lambda x: inside_bounds(bounds, x), x)))


def bounds_range(bounds_tuple):
    """Returns the range of a single bound
    Parameters
    ----------
    bounds_tuple : tuple object

    Returns
    -------
    numeric
    """
    return np.abs(bounds_tuple[1] - bounds_tuple[0])


def bounds_range_ndim(bounds):
    """Returns the range of a single bound
    Parameters
    ----------
    bounds : list of tuples
        Each tuple in the list defines the bounds for the corresponding variable
        Example: [(1, 2), (2, 3), (-1, 4)...]

    Returns
    -------
    array-like of shape (len(bounds),)
    """
    return np.array(list(map(bounds_range, bounds)))


# Python Standard Libraries
import warnings
import time
import os
import sys
from pathlib import Path

# Third party imports
# fancy prints
import numpy as np
from tqdm import tqdm

# grAdapt package
import grAdapt.utils.math
import grAdapt.utils.misc
import grAdapt.utils.sampling
from grAdapt import surrogate as sur, optimizer as opt, escape as esc
from grAdapt.space.transformer import Transformer
from grAdapt.sampling import initializer as init, equidistributed as equi


class Sequential:
    def __init__(self, surrogate=None, optimizer=None, sampling_method=None,
                 initializer=None, escape=None,
                 training=None, random_state=1):
        """
        Parameters
        ----------
        surrogate : grAdapt Surrogate object
        optimizer : grAdapt Optimizer object
        sampling_method : Sampling Method to be used. static method from utils
        initializer : grAdapt Initializer object
        escape : grAdapt Escape object
        training : (X, y) with X shape (n, m) and y shape (n,)
        random_state : integer
            random_state integer sets numpy seed
        """

        # Standard Values
        if surrogate is None:
            self.surrogate = sur.GPRSlidingWindow()
        else:
            self.surrogate = surrogate

        if optimizer is None:
            self.optimizer = opt.AMSGradBisection(surrogate=self.surrogate)
        else:
            self.optimizer = optimizer
            if surrogate is None:
                raise Exception('If optimizer is passed, then surrogate must be passed, too.')

        if sampling_method is None:
            self.sampling_method = equi.MaximalMinDistance()
        else:
            self.sampling_method = sampling_method

        if initializer is None:
            self.initializer = init.VerticesForceRandom(sampling_method=self.sampling_method)
        else:
            self.initializer = initializer
            if sampling_method is None:
                raise Exception('If initializer is passed, then sampling_method must be passed, too.')

        if escape is None:
            self.escape = esc.NormalDistributionDecay(surrogate=self.surrogate, sampling_method=self.sampling_method)
        else:
            self.escape = escape
            if surrogate is None or sampling_method is None:
                raise Exception('When passing an escape function, surrogate and sampling_method must be passed, too.')

        # continue optimizing
        self.training = training
        if training is not None:
            self.X = training[0]
            self.y = training[1]
            # self.fit(self.X, self.y)
        else:
            self.X = None
            self.y = None

        # seed
        self.random_state = random_state
        np.random.seed(self.random_state)

        # results
        self.res = None

        # keep track of checkpoint files
        self.checkpoint_file = None
        self.auto_checkpoint = False

    def warn_n_evals(self):
        if self.n_evals <= 0:
            raise Exception('Please set n_evals higher higher than 0.')

        if self.n_evals <= self.dim:
            warnings.warn('n_evals should be higher than the dimension of the problem.')

        if self.n_random_starts == 'auto' or not isinstance(self.n_random_starts, (int, float)):
            self.n_random_starts = grAdapt.utils.misc.random_starts(self.n_evals, self.dim)

        if self.n_evals < self.n_random_starts:
            warnings.warn('n_random starts can\'t be higher than n_evals.')
            warnings.warn('n_random_starts set automatically. ')
            self.n_random_starts = grAdapt.utils.misc.random_starts(self.n_evals, self.dim)

    def fit(self, X, y):
        """Fit known points on surrogate model

        Parameters
        ----------
        X : array-like (n, m)
        y : array (n,)

        Returns
        -------
        None

        """
        print('Surrogate model fitting on known points.')
        self.surrogate.fit(X, y)

    def escape_x_criteria(self, x_train, iteration):
        """Checks whether new point is different than the latest point by the euclidean distance
        Checks whether new point is inside the defined search space/bounds.
        Returns True if one of the conditions above are fulfilled.

        Parameters
        ----------
        x_train : ndarray (n, d)
        iteration : integer

        Returns
        -------
        boolean
        """
        # x convergence
        escape_convergence = (np.linalg.norm(x_train[iteration - 1] - x_train[iteration])) < self.eps
        # check whether point is inside bounds
        escape_valid = not (grAdapt.utils.sampling.inside_bounds(self.bounds, x_train[iteration]))

        escape_x = escape_convergence or escape_valid
        return escape_x

    @staticmethod
    def escape_y_criteria(y_train, iteration, pct):
        """

        Parameters
        ----------
        y_train : array-like (n, d)
        iteration : integer
        pct : numeric
            pct should be less than 1.

        Returns
        -------
        boolean
        """
        try:
            return grAdapt.utils.misc.is_inside_relative_range(y_train[iteration - 1], y_train[iteration - 2], pct)
        except:
            return False

    def minimize(self, func, bounds, n_evals='auto', eps=1e-3, f_min=-np.inf, f_min_eps=1e-2, n_random_starts='auto',
                 auto_checkpoint=False, show_progressbar=True, prints=True):
        """

        Parameters
        ----------
        func : takes ndarray and returns scalar
        bounds : list
             list of tuples e.g. [(-5, 5), (-5, 5)]
        n_evals : int or string
            number of max. function evaluations
            'auto' : will evaluate func based on the probability of
                    hitting the optimal solution by coincidence
                    between 100 and 10000
        eps : float
            convergence criteria, absolute tolerance
        f_min : numeric
            if the minimal target value of func is known
            can lead to earlier convergence
        f_min_eps : numeric
            early stop criteria, relative tolerance
        n_random_starts : string or int
        auto_checkpoint : bool
        show_progressbar : bool
        prints : bool

        Returns
        -------
        res: dictionary
            res['x'] : ndarray (n, d)
            res['y'] : ndarray (n,)
            res['x_sol'] : solution vector
            res['y_sol'] : y solution
            res['surrogate'] : grAdapt surrogate object

        """
        # print('Optimizing ' + func.__name__)
        self.func = Transformer(func, bounds)
        #self.bounds = List()  # Numba Typed List for performance
        #[self.bounds.append((float(x[0]), float(x[1]))) for x in bounds]
        self.bounds = bounds
        self.n_evals = int(n_evals)
        self.eps = eps
        self.f_min = f_min
        self.f_min_eps = f_min_eps
        self.n_random_starts = n_random_starts
        self.auto_checkpoint = auto_checkpoint
        self.dim = len(bounds)
        self.prints = prints

        if self.prints is False:
            sys.stdout = open(os.devnull, 'w')
        """n_evals value based on the probability of finding
        the optimal solution by coincidence
        """
        if self.n_evals == 'auto':
            vol_sphere = grAdapt.utils.math.geometry.volume_hypersphere(len(bounds), self.eps)
            vol_rec = grAdapt.utils.math.geometry.volume_hyperrectangle(bounds)

            # limit n_evals to 100 and 10000
            self.n_evals = max(100, int(vol_rec / vol_sphere))
            self.n_evals = min(10000, self.n_evals)

        """Catching errors/displaying warnings related to n_evals and n_random_starts
        and automatically set n_random_starts if not given
        """
        self.warn_n_evals()

        """checkpoint print directory
        """
        if auto_checkpoint:
            directory_path = os.getcwd() + '/checkpoints'
            print('auto_checkpoint set to True. The training directory is located at\n' + directory_path)

        """Inittialize x_train and y_train
        Check whether training can be continued
        """
        # TODO: Change number of dimensions for nominal datatype
        x_train = np.zeros((self.n_evals, self.dim))
        y_train = np.zeros((self.n_evals,)) + np.inf
        if self.training is not None:
            x_train = np.vstack((self.X, x_train))
            y_train = np.hstack((self.y, y_train))

            # prevent same sample points in training continuation
            self.random_state += 1
            np.random.seed(self.random_state)
            print('Training data added successfully.')

        """Randomly guess n_random_starts points
        """
        print('Sampling {0} random points.'.format(self.n_random_starts))
        print('Random function evaluations. This might take a while.')
        train_len = len(y_train) - self.n_evals

        if train_len == 0:
            x_train[train_len:self.n_random_starts + train_len] = \
                self.initializer.sample(self.bounds, self.n_random_starts)
            y_train[train_len:self.n_random_starts + train_len] = \
                np.array(list(map(self.func, tqdm(x_train[train_len:self.n_random_starts + train_len],
                                                  total=self.n_evals, leave=False))))
        else:
            x_train[train_len:self.n_random_starts + train_len] = \
                self.sampling_method.sample(self.bounds, self.n_random_starts, x_history=x_train[:train_len])
            y_train[train_len:self.n_random_starts + train_len] = \
                np.array(list(map(self.func, tqdm(x_train[train_len:self.n_random_starts + train_len],
                                                  total=self.n_evals, leave=False))))

        print('Finding optimum...')

        """Start from best point
        """

        best_idx = np.argmin(y_train[:self.n_random_starts + train_len])

        # swap positions
        x_train[[best_idx, self.n_random_starts - 1 + train_len]] = \
            x_train[[self.n_random_starts - 1 + train_len, best_idx]]
        y_train[[best_idx, self.n_random_starts - 1 + train_len]] = \
            y_train[[self.n_random_starts - 1 + train_len, best_idx]]

        """Optimizing loop
        """
        start_time = time.perf_counter()
        pbar = tqdm(total=self.n_evals + train_len, initial=self.n_random_starts + train_len,
                    disable=not show_progressbar)
        for iteration in range(self.n_random_starts + train_len, self.n_evals + train_len):
            # print(iteration)
            pbar.update()  # progressbar update

            # Fit data on surrogate model
            self.surrogate.fit(x_train[:iteration], y_train[:iteration])

            # gradient parameters specific for the surrogate model
            surrogate_grad_params = [x_train[:iteration], y_train[:iteration], self.func, bounds]
            # print(x_train[iteration-1])
            x_train[iteration] = self.optimizer.run(x_train[iteration - 1], grAdapt.utils.misc.epochs(iteration),
                                                    surrogate_grad_params)

            escape_x_criteria_boolean = self.escape_x_criteria(x_train, iteration)
            escape_y_criteria_boolean = self.escape_y_criteria(y_train, iteration, self.f_min_eps)
            escape_boolean = escape_x_criteria_boolean or escape_y_criteria_boolean

            # sample new point if must escape or bounds not valid
            if escape_boolean:
                x_train[iteration] = self.escape.get_point(x_train[:iteration], y_train[:iteration], iteration,
                                                           self.bounds)

            # obtain y_train
            y_train[iteration] = self.func(x_train[iteration])

            # stop early

            if grAdapt.utils.misc.is_inside_relative_range(y_train[iteration], self.f_min, self.f_min_eps):
                break

            # auto_checkpoint
            if auto_checkpoint and time.perf_counter() - start_time >= 60:
                self.X = x_train
                self.y = y_train
                res = self.build_res()
                self.save_checkpoint(res)
                start_time = time.perf_counter()

        # progressbar
        pbar.close()

        self.X = x_train
        self.y = y_train
        # save current training data
        self.training = (self.X, self.y)
        self.res = self.build_res()

        if auto_checkpoint:
            self.save_checkpoint(self.res)

        # restore prints
        if self.prints is False:
            sys.stdout = sys.__stdout__

        return self.res

    def maximize(self, func, bounds, *args, **kwargs):
        """

        Parameters
        ----------
        func : takes ndarray and returns scalar
        bounds : list of tuples e.g. [(-5, 5), (-5, 5)]
        args : args from minimize
        kwargs : args from minimize

        Returns
        -------
        res: dictionary
            res['x'] : ndarray (n, d)
            res['y'] : ndarray (n,)
            res['x_sol'] : solution vector
            res['y_sol'] : y solution
            res['surrogate'] : grAdapt surrogate object
        """

        def f_max(x):
            return -func(x)

        # x_train, y_train, surrogate = self.minimize(f_max, bounds, *args, **kwargs)
        res = self.minimize(f_max, bounds, *args, **kwargs)
        res['y'] = -res['y']
        res['y_sol'] = -res['y_sol']

        # save the right y values
        if self.auto_checkpoint:
            self.save_checkpoint(res)

        return res

    def minimize_args(self, func, bounds, *args, **kwargs):
        """

        Parameters
        ----------
        func : takes ndarray and returns scalar
        bounds : list of tuples e.g. [(-5, 5), (-5, 5)]
        args : args from minimize
        kwargs : args from minimize

        Returns
        -------
        res: dictionary
            res['x'] : ndarray (n, d)
            res['y'] : ndarray (n,)
            res['x_sol'] : solution vector
            res['y_sol'] : y solution
            res['surrogate'] : grAdapt surrogate object
        """

        def f_min_args(*args2):
            arr = args2[0]
            args_list = arr.tolist()

            return func(*args_list)

        # x_train, y_train, surrogate = self.minimize(f_max, bounds, *args, **kwargs)
        res = self.minimize(f_min_args, bounds, *args, **kwargs)

        return res

    def maximize_args(self, func, bounds, *args, **kwargs):
        """

        Parameters
        ----------
        func : takes ndarray and returns scalar
        bounds : list of tuples e.g. [(-5, 5), (-5, 5)]
        args : args from minimize
        kwargs : args from minimize

        Returns
        -------
        res: dictionary
            res['x'] : ndarray (n, d)
            res['y'] : ndarray (n,)
            res['x_sol'] : solution vector
            res['y_sol'] : y solution
            res['surrogate'] : grAdapt surrogate object
        """

        def f_max(*args2):
            return -func(*args2)

        # x_train, y_train, surrogate = self.minimize(f_max, bounds, *args, **kwargs)
        res = self.minimize_args(f_max, bounds, *args, **kwargs)
        res['y'] = -res['y']
        res['y_sol'] = -res['y_sol']

        # save the right y values
        if self.auto_checkpoint:
            self.save_checkpoint(res)

        return res

    def scipy_wrapper_minimize(self, func, x0, bounds=None, tol=None, *args, **kwargs):
        """
        @ https://github.com/scipy/scipy/blob/v1.5.3/scipy/optimize/_minimize.py
        Parameters
        ----------
        func : callable
            The objective function to be minimized.
            ``fun(x, *args) -> float``
            where ``x`` is an 1-D array with shape (d,) and ``args``
            is a tuple of the fixed parameters needed to completely
            specify the function.
        x0 : ndarray, shape (d,)
            Initial guess. Array of real elements of size (d,),
        bounds : sequence or `Bounds`, optional
            There are two ways to specify the bounds:

                1. Instance of `Bounds` class.
                2. Sequence of ``(min, max)`` pairs for each element in `x`. None
                   is used to specify no bound.
        tol : float, optional
            Tolerance for termination. For detailed control, use solver-specific options.
        -------

        Returns
        -------
        ndarray (d,)
        """
        self.training = (x0.reshape(1, -1), func(x0))
        self.X = self.training[0]
        self.y = self.training[1]
        # self.fit(self.X, self.y)
        res = self.minimize(func, bounds, *args, **kwargs)

        return res['x_sol']

    def build_res(self):
        # define output
        transformed_input = []
        for i in range(self.X.shape[0]):
            arguments = np.zeros((self.X.shape[1],), dtype='O')
            for j in range(len(arguments)):
                try:
                    arguments[j] = self.bounds[j].transform(self.X[i, j])
                except:
                    arguments[j] = self.X[i, j]

            transformed_input.append(arguments)

        res = {'x': np.array(transformed_input), 'x_internal': self.X, 'y': self.y,
               'x_sol': np.array(transformed_input)[np.argmin(self.y)],
               'x_sol_internal': self.X[np.argmin(self.y)], 'y_sol': np.min(self.y),
               'surrogate': self.surrogate,
               'optimizer': self.optimizer}

        return res

    def load_checkpoint(self, filename):
        """

        Parameters
        ----------
        filename : string to filepath of checkpoint

        Returns
        -------
        None

        """
        res = np.load(filename, allow_pickle=True).item()

        self.X = res['x']
        self.y = res['y']

        return res
        # self.surrogate = res['surrogate']

    def save_checkpoint(self, res):
        """

        Parameters
        ----------
        res: dictionary
            res['x'] : ndarray (n, d)
            res['y'] : ndarray (n,)
            res['x_sol'] : solution vector
            res['y_sol'] : y solution
            res['surrogate'] : grAdapt surrogate object

        Returns
        -------
        None

        """
        directory = Path(os.getcwd()) / 'checkpoints'
        filename = ('checkpointXY' + time.strftime('%y%b%d-%H%M%S') + '.npy')
        filename = directory / filename

        if not os.path.exists(directory):
            os.makedirs(directory)

        # save new file
        np.save(filename, res)
        print('Checkpoint created in ' + str(filename))

        # delete last file
        if self.checkpoint_file is not None:
            os.remove(self.checkpoint_file)
            print('Old checkpoint file {0} deleted.'.format(self.checkpoint_file))
        self.checkpoint_file = filename



import unittest
from grAdapt.models import Sequential

import numpy as np


class TestEqualYValues(unittest.TestCase):

    @staticmethod
    def rastrigin(x):
        x = np.array(x)
        return 10 * len(x) + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x), axis=0)

    @staticmethod
    def sphere(x):
        return np.sum(x ** 2)

    def test_equal_sphere(self):
        model = Sequential()
        bounds = [(-10, 10) for _ in range(5)]
        n = 50
        res = model.minimize(self.sphere, bounds, n, show_progressbar=False)
        x_train = res['x']
        y_train = res['y']

        y_compare = np.array(list(map(self.sphere, x_train)))

        self.assertTrue(np.array_equal(y_compare, y_train))

    def test_equal_rastrigin(self):
        model = Sequential()
        bounds = [(-10, 10) for _ in range(5)]
        n = 50
        res = model.minimize(self.rastrigin, bounds, n, show_progressbar=False)
        x_train = res['x']
        y_train = res['y']

        y_compare = np.array(list(map(self.rastrigin, x_train)))

        self.assertTrue(np.array_equal(y_compare, y_train))


if __name__ == '__main__':
    unittest.main()



# python
import warnings

# Third party imports
import numpy as np

# grAdapt
from .base import Initial
from grAdapt.utils.sampling import sample_corner_bounds


class VerticesForceRandom(Initial):
    """
    Samples all vertices if n_evals >= 2 ** len(bounds).
    Else, a subset of vertices is sampled.
    """

    def __init__(self, sampling_method):
        """
        Parameters
        ----------
        sampling_method : grAdapt.sampling.equidistributed Object
            Sample low discrepancy sequences when initial point method is not feasible
        """
        super().__init__(sampling_method)

    def sample(self, bounds, n_evals):
        """Returns a numpy array of sampled points.
        Does not include corner points of the hypercube/search space.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n_evals : int
            number of initial points sampled by method

        Returns
        -------
        (self.n_evals, len(self.bounds)) numpy array
        """
        super().sample(bounds, n_evals)

        if 2 ** len(self.bounds) >= self.n_evals:
            # sample corner points first which fits in n_evals
            d_tilde = int(np.floor(np.log2(self.n_evals)))
            corners_d_tilde = sample_corner_bounds(self.bounds[:d_tilde])  # (2 ** d_tilde, d_tilde) array
            n_tilde = 2 ** d_tilde
            # sample random fixed corner points
            random_binary_array = np.random.randint(2, size=(len(self.bounds),))
            remainder_bounds = self.bounds[d_tilde:]
            fix_corners = np.zeros((1, len(remainder_bounds)))
            for i in range(len(remainder_bounds)):
                if random_binary_array[i] == 0:
                    fix_corners[0][i] = remainder_bounds[i][0]
                else:
                    fix_corners[0][i] = remainder_bounds[i][1]
            fix_corners_2d = np.tile(fix_corners, (n_tilde, 1))  # (n, d-d_tilde)

            # corner points with fixed rest dimensions
            corner_points_fixed = np.hstack((corners_d_tilde, fix_corners_2d))

            # because 2 ** n_tilde <= n, sample n - n_tilde
            if self.n_evals - n_tilde > 0:
                random_points = self.sampling_method.sample(bounds=self.bounds,
                                                            n=(self.n_evals - n_tilde),
                                                            x_history=corner_points_fixed)
                return np.vstack((corner_points_fixed, random_points))
            else:
                return corner_points_fixed

        else:
            corner_points = sample_corner_bounds(self.bounds)
            num_corner_points = corner_points.shape[0]
            random_points = self.sampling_method.sample(bounds=self.bounds,
                                                        n=(self.n_evals - num_corner_points),
                                                        x_history=corner_points)

            return np.vstack((corner_points, random_points))



# Python Standard Libraries
import warnings

# grAdapt
# from .base import Equidistributed
from .MaximalMinDistance import MaximalMinDistance


class Mitchell(MaximalMinDistance):
    """
    [Mitchell et al., 1991],
    Spectrally optimal sampling for distribution ray tracing
    """
    def __init__(self, m=3):
        """
        Parameters
        ----------
        m : integer
            number of candidates = m * n
        """
        warnings.warn('Mitchell\' best candidate has a time complexity of O(n^3) '
                      'and memory issues when dealing with higher sample numbers. '
                      'Use MaximalMinDistance instead which is an improved version '
                      'with linear time complexity.', ResourceWarning)
        super().__init__(n_candidates=m, window_size=0)
        self.candidates_set = False

    def sample(self, bounds, n, x_history=None):
        """Samples low discrepancy/equidistributed sequences according to Mitchell.
        Method has to handle with new bounds and n.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n : int
            number of points to be sampled
        x_history : array-like (2d)
            History points. Consider those to prevent sampling in dense regions.

        Returns
        -------
        array-like (n, len(bounds))
            Returns a 2D array. dim is the dimension of a single point
            Each row corresponds to a single point.
            Each column corresponds to a dimension.
        """
        if self.candidates_set is False:
            self.n_candidates = self.n_candidates * n
            self.candidates_set = True
        return super().sample(bounds, n, x_history)



from grAdapt.models import Sequential
import numpy as np
from utils import timer, enablePrint

def sphereMax(x):
    return -np.sum(x**2)
    
def sphereMin(x):
    return np.sum(x**2)

@timer
def test1():
    model = Sequential()
    bounds = [(-10, 10) for i in range(5)]
    res = model.maximize(sphereMax, bounds, 20, show_progressbar=False)
    x = res['x']
    y = res['y']
    print('Maximum found: {ymax}'.format(ymax=np.max(y)))
    print('Argmax: {xmax}'.format(xmax=x[np.argmax(y)]))
    
    return np.max(y)

@timer
def test2():
    model = Sequential()
    bounds = [(-10, 10) for i in range(5)]
    res = model.minimize(sphereMin, bounds, 20, show_progressbar=False)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    print('Argmin: {xmin}'.format(xmin=x[np.argmin(y)]))
    return np.min(y)



def main():
    x1 = test1()
    x2 = test2()
    try:
        assert x1 == -x2
        print('Maximize \t \t \t Ok.')
    except:
        enablePrint()
        print('Maximize \t \t \t Not Ok.')
main()


import unittest
import grAdapt
from grAdapt.sampling.initializer import Standard, Vertices, VerticesForce, VerticesForceRandom
from grAdapt.sampling.equidistributed import MaximalMinDistance
from grAdapt.utils.sampling import inside_bounds_ndim


class TestInitalizer(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        sampling_method = MaximalMinDistance()
        self.all_initializers = [Standard(sampling_method),
                                 Vertices(sampling_method),
                                 VerticesForce(sampling_method),
                                 VerticesForceRandom(sampling_method)]
        super().__init__(*args, **kwargs)

    def test_sample_number(self):
        """
        Test all Initializers in all_initializers.
        Tests for several dimension and sample sizes
        whether the first-axis of the output equals
        the sample size n.
        """
        # sampling_method = MaximalMinDistance()
        # all_initializers = [Standard(sampling_method), Vertices(sampling_method),
        #                   VerticesForce(sampling_method)]

        n_samples = 33
        dim = 20

        for initializer in self.all_initializers:
            for i in range(1, n_samples):
                for d in range(1, dim):
                    bounds = [(-10, 10) for i in range(d)]
                    sample = initializer.sample(bounds, i)
                    self.assertEqual(sample.shape[0], i)


if __name__ == '__main__':
    unittest.main()



# Third party imports
import numpy as np
from sklearn.gaussian_process.kernels import RBF as skRBF, RationalQuadratic as skRationalQuadratic

# grAdapt
from grAdapt.utils.math.linalg import inv_stable


class Nystroem:
    """Formula derived by William and Seeger
    Using the Nyström Method to Speed Up Kernel Machines (2001)

    Nystroem Method: avoid calculating all kernel entries
    Kernel is passed uninitialized.

    Nystroem can be used to approximate the Gram Matrix:
    >>> from grAdapt.surrogate.kernels import Nystroem, RationalQuadratic
    >>> import numpy as np
    >>> xn = np.linspace(-10, 10)
    >>> kernel = RationalQuadratic()
    >>> gram_exact = kernel(xn)
    >>> kernel_nystroem = Nystroem(kernel)
    >>> gram_approx = kernel_nystroem(xn)
    >>> print(np.linalg.norm(gram_exact-gram_approx))

    Nystroem kernel can be passed to the surrogate just like any kernel:
    >>> from grAdapt.surrogate.kernels import Nystroem, RBF
    >>> from grAdapt.surrogate import GPRSlidingWindow
    >>> from grAdapt.models import Sequential
    >>> rbf = RBF()
    >>> kernel = Nystroem(rbf)
    >>> gpr = GPRSlidingWindow(kernel=Nystroem, window_size = 10)
    >>> model = Sequential(surrogate=gpr)
    """

    def __init__(self, kernel, n_components=100):
        """

        Parameters
        ----------
        kernel : kernel initialized object
        n_components : int
            number of rows to be evaluated
        # kernel_params : kernel parameters e.g. length_scale, alpha etc
        """
        self.kernel = kernel
        self.n_components = n_components
        self.kernel_name = self.kernel.__str__()

        if hasattr(self.kernel, 'alpha'):
            self.alpha = self.kernel.alpha
        if hasattr(self.kernel, 'length_scale'):
            self.length_scale = self.kernel.length_scale

    def __call__(self, X1, X2=None):
        """

        Parameters
        ----------
        X1 : array-like (n, d)
        X2 : array-like (n, d)

        Returns
        -------
        array-like (n, n)
            Kernel Matrix
        """

        if len(X1.shape) == 1:
            X1 = X1.reshape(1, -1)

        if X2 is None:
            X2 = X1

        n = X1.shape[0]
        k = self.n_components

        if k < n:
            AC = self.kernel(X1[:k], X2)
            A = AC[:, :k]
            C = AC[:, k:].T

            # Cholesky inverse for numerical stability
            B = C @ inv_stable(A) @ C.T
            CB = np.hstack((C, B))

            del B
            del C
            del A

            return np.vstack((AC, CB))
        else:
            return self.kernel(X1, X2)

    def get_params(self, deep=True):
        return self.kernel.get_params(deep)


def set_lengthscale(bounds):
    """

    Parameters
    ----------
    bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]

    Returns
    -------

    """
    return np.sqrt(np.sum([(np.abs(u-l)/3)**2 for (l, u) in bounds]))


class RBF(skRBF):
    """Radial Basis Function
    Modified length scale to counter curse of dimensionality.
    """
    def __init__(self, bounds=None):
        super().__init__()
        if bounds is not None:
            self.length_scale = set_lengthscale(bounds)

        self.__name__ = 'RBF'

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__class__.__name__

    def __call__(self, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.
         @ docstring by scikit-learn
         https://github.com/scikit-learn/scikit-learn/blob/0fb307bf3/sklearn/gaussian_process/kernels.py#L1315

        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        Returns
        -------
        K : ndarray of shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : ndarray of shape (n_samples_X, n_samples_X, n_dims), \
                optional
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        if self.length_scale == 1.0:
            self.length_scale = np.sqrt(X.shape[1])

        return super().__call__(X, Y, eval_gradient)

    def derivative(self):
        pass


class RationalQuadratic(skRationalQuadratic):
    """Rational Quadratic function
     Modified length scale to counter curse of dimensionality.
    """
    def __init__(self, bounds=None):
        super().__init__()
        if bounds is not None:
            self.length_scale = set_lengthscale(bounds)

        self.__name__ = 'RationalQuadratic'

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__class__.__name__

    def __call__(self, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.
         @ docstring by scikit-learn
         https://github.com/scikit-learn/scikit-learn/blob/0fb307bf3/sklearn/gaussian_process/kernels.py#L1315

        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        Returns
        -------
        K : ndarray of shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : ndarray of shape (n_samples_X, n_samples_X, n_dims), \
                optional
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        if self.length_scale == 1.0:
            self.length_scale = np.sqrt(X.shape[1])

        return super().__call__(X, Y, eval_gradient)

    def derivative(self):
        pass


# Python Standard Libraries
# decorators
from abc import abstractmethod


class Surrogate:
    """Surrogate model object for the regression
       
       Parameters
       ----------
       model : Regression model object
            Can be a scikit-learn object. It must support
                .fit(X, y)
                .predict(X, return_std) (maybe?)
            
       surrogate_params : list or tuple
            parameters passed for the surrogate_model
       

       Attributes
       ----------
       self.model : model
       self.surrogate_params : surrogate_params


       Examples
       --------
       >>> surrogate = Surrogate(surrogate_params)
       >>> surrogate.fit(X, y)
       >>> surrogate.eval_gradient(x, surrogate_grad_params)
    """

    def __init__(self, model, surrogate_params):
        self.model = model
        self.surrogate_params = surrogate_params

    @abstractmethod
    def fit(self, X, y):
        """ Fits x points to the y-coordinate
        
            Parameters
            ----------
            X : array-like of size (n, m) for n points of d dimensions
            y : array-like of shape (n_samples,)
            
            
            Returns
            -------
            None
        """
        # self.model.fit(X, y)
        raise NotImplementedError

    @abstractmethod
    def predict(self, X, **args):
        return NotImplementedError

    @abstractmethod
    def eval_gradient(self, x, surrogate_grad_params):
        """Evaluates the gradient of the surrogate model at point x
        
        Parameters
        ----------
        x : 1D array-like
        
        
        Returns
        -------
        grad : 1D array-like
            same shape as x
        """

        raise NotImplementedError



# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Equidistributed
from grAdapt.utils.sampling import sample_points_bounds
from grAdapt.utils.math.spatial import pairwise_distances


class KDistance(Equidistributed):
    """K-Distance sampling method

    Candidates are firstly generated. The candidate with the largest k-distance is sampled.
    """

    def __init__(self, n_candidates=10, k=3, window_size=500):
        """
        Parameters
        ----------
        k : integer
            k-th element from a point
        window_size : integer
            size of history points to consider. smaller is faster but worsen the results.
        """
        super().__init__()
        self.n_candidates = n_candidates
        self.k = k
        self.window_size = window_size

    def sample(self, bounds, n, x_history=None):
        """Samples low discrepancy/equidistributed sequences
        Method has to handle with new bounds and n.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n : int
            number of points to be sampled
        x_history : array-like (2d)
            History points. Consider those to prevent sampling in dense regions.

        Returns
        -------
        array-like (n, len(bounds))
            Returns a 2D array. dim is the dimension of a single point
            Each row corresponds to a single point.
            Each column corresponds to a dimension.
        """

        # set to new variables
        super().sample(bounds, n, x_history)

        if x_history is None:
            x_history = sample_points_bounds(self.bounds, 1)
            x_history_list = list(x_history)
        else:
            x_history_list = list(x_history)

        for i in range(self.n):
            x_history_sublist = x_history_list[-self.window_size:]
            candidates = sample_points_bounds(self.bounds, self.n_candidates)
            dists_matrix = pairwise_distances(candidates, np.array(x_history_sublist))

            # k-distance
            sorted_dists = np.sort(dists_matrix, axis=1)
            k_dists = sorted_dists[:, min(self.k-1, i)] # if samples too low
            best_candidate_idx = np.argmax(k_dists)
            x_history_list.append(candidates[best_candidate_idx])

        return np.array(x_history_list)[-self.n:]








# Python Standard Libraries
import numpy as np

# grAdapt package
from .base import Escape, inside_bounds, sample_points_bounds, bounds_range_ndim


class MaxStandardDeviation(Escape):

    def __init__(self, surrogate, sampling_method=None):
        super().__init__(surrogate, sampling_method)

    def get_point(self, x_train, y_train, iteration, bounds):
        # ignore warnings numerical issues
        import warnings
        warnings.filterwarnings('ignore')

        x_best = x_train[np.argmin(y_train)]  # x_best is current optimizer and "trust region"
        bounds_vec = ((bounds_range_ndim(bounds)/6)**2) / (np.log(iteration + np.e))  # shrink region to discover
        cov_matrix = np.diag(bounds_vec)

        # TODO: Points outside of bounds resample uniformly
        best_point = None
        best_std = 0

        # Many points created near current best x point
        # Pick the one where the standard deviation is highest
        for i in range(20):

            current_point = np.random.multivariate_normal(x_best, cov_matrix)  # (d, )
            _, current_std = self.surrogate.predict(current_point.reshape(1, -1), return_std=True)

            # max std and if is inside bounds
            if current_std >= best_std and inside_bounds(bounds, current_point):
                best_std = current_std
                best_point = current_point

        if best_point is None:
            return self.escape_history(bounds, x_train)
        else:
            return best_point




import numpy as np
from scipy.special import gamma

# package
from grAdapt.utils.sampling import bounds_range


def volume_hypersphere(dim, radius=1):
    """Calculates the volume of a hypersphere given the radius and the dimension

    Parameters
    ----------
    dim : int
    radius : float

    Returns
    -------
    float
    """
    return (radius ** dim) * np.pi**(dim/2) / (gamma(dim/2 + 1))


def volume_hyperrectangle(bounds):
    """Calculates the volume of a hyperrectangle of any dimension

    Parameters
    ----------
    bounds : list of tuples.
        Each tuple in the list defines the bounds for the corresponding variable
        Example: [(1, 2), (2, 3), (-1, 4)...]

    Returns
    -------

    """
    return np.prod(list(map(bounds_range, bounds)))



# https://en.wikipedia.org/wiki/Test_functions_for_optimization

import numpy as np

# Rastrigin

def rastrigin(x):
    x = np.array(x)
    return 10 * len(x) + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x), axis=0)


def rastrigin_bounds(d):
    """

    Parameters
    ----------
    d : int
        dimension

    Returns
    -------
    list of 2-tuples
    """
    return [(-5.12, 5.12) for _ in range(d)]


def rastrigin_ySol(d):
    return 0


def rastrigin_xSol(d):
    """

    Parameters
    ----------
    d : int
        dimension

    Returns
    -------

    """
    return np.zeros((d,))

# Sphere

def sphere(x):
    x = np.array(x)
    return np.sum(x ** 2)


def sphere_bounds(d):
    """

    Parameters
    ----------
    d : int
        dimension

    Returns
    -------

    """
    return [(-10., 10.) for _ in range(d)]


def sphere_ySol(d):
    return 0


def sphere_xSol(d):
    """

    Parameters
    ----------
    d : int
        dimension

    Returns
    -------

    """
    return np.zeros((d,))


# Rosenbrock

def rosenbrock(x):
    x = np.array(x)
    summation = 0
    for i in range(len(x)-1):
        summation += 100*((x[i+1] - x[i]**2)**2) + (1 - x[i])**2

    return summation


def rosenbrock_bounds(d):
    """

    Parameters
    ----------
    d : int
        dimension

    Returns
    -------

    """
    return [(-10., 10.) for i in range(d)]


def rosenbrock_ySol(d):
    return 0


def rosenbrock_xSol(d):
    return np.ones((d,))


# Styblinski

def styblinski(x):
    x = np.array(x)
    return np.sum(x ** 4 - 16 * x ** 2 + 5 * x) / 2


def styblinski_bounds(d):
    """

    Parameters
    ----------
    d : int
        dimension

    Returns
    -------

    """
    return [(-5., 5.) for _ in range(d)]


def styblinski_ySol(d):
    return -39.16616 * d


def styblinski_xSol(d):
    return np.ones((d,)) * -2.903534

all_functions_string = ['rastrigin', 'sphere', 'rosenbrock', 'styblinski']

all_functions = [eval(function_string) for function_string in all_functions_string]

all_bounds = [eval(function_string+'_bounds') for function_string in all_functions_string]

all_xSol = [eval(function_string+'_xSol') for function_string in all_functions_string]

all_ySol = [eval(function_string+'_ySol') for function_string in all_functions_string]


from grAdapt.models import Sequential
import numpy as np
from utils import timer, enablePrint
from grAdapt.surrogate import GPR, GPRSlidingWindow, GPROnlineInsert
from grAdapt.surrogate.kernels import Nystroem, RationalQuadratic, RBF

def rastrigin(x):
    x = np.array(x)
    return 10*len(x)+np.sum(x**2-10*np.cos(2*np.pi*x), axis=0)
    
def sphereMin(x):
    return np.sum((x-1.25)**2)

@timer
def test():
    gpr = GPRSlidingWindow(kernel=Nystroem(RBF()), window_size = 10)
    model = Sequential(surrogate=gpr)
    bounds = [(-10, 10) for i in range(11)]
    res = model.minimize(sphereMin, bounds, 100, show_progressbar=False)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    print('Argmin: {xmin}'.format(xmin=x[np.argmin(y)]))
    
def main():
    try:
        test()
        print('Sliding Window \t \t \t Ok.')
    except:
        enablePrint()
        print('Sliding Window \t \t \t Not Ok.')
main()


# Third party imports
import numpy as np

# Package imports
from .base import Surrogate


class NoGradient(Surrogate):
    """NoGradient Surrogate
    Returns a gradient of zero entries
    Optimizer will converge and Model will exploit
    """

    def __init__(self, **surrogate_params):
        super().__init__(None, surrogate_params)

    def fit(self, X, y):
        """Fit training data (X, y)
        Parameters
        ----------
        X : array-like (n, d)
        y : array-like (n,)


        Returns
        -------
        None

        """
        pass

    def predict(self, x_predict, return_std=False):
        """Predict y value based on x_predict
        Parameters
        ----------
        x_predict : array-like (n, d)
        return_std : array-like (n,)
            returns the standard deviation of the posterior mean


        Returns
        -------
        array-like (n,)

        """
        if return_std:
            return np.zeros_like(x_predict), np.zeros_like(x_predict)
        else:
            return np.zeros_like(x_predict)

    def eval_gradient(self, x, surrogate_grad_params):
        """Does not evaluate the gradient
        Parameters
        ----------
        x: 1D array-like (d,)
        surrogate_grad_params: list
            surrogate_grad_params[0]: whole training data shape (m, d)


        Returns
        -------
        grad : np.zeros_like(x)
            same shape as x
        """
        return np.zeros_like(x)



# Python Standard Libraries
# import numpy as np
# decorators
from abc import abstractmethod

# grAdapt package
from grAdapt.utils.sampling import *
from grAdapt.sampling.equidistributed import MaximalMinDistance


class Escape:
    """Escape Base class

    Parameters
    ----------
    self.surrogate : surrogate object

    Attributes
    ----------


    Examples
    --------
    """

    def __init__(self, surrogate, sampling_method=None):
        self.surrogate = surrogate
        if sampling_method is None:
            self.sampling_method = MaximalMinDistance()
        else:
            self.sampling_method = sampling_method

    def __call__(self, surrogate):
        self.surrogate = surrogate

    def escape_history(self, bounds, x_train):
        """Escape considering history of points

        Parameters
        ----------
        bounds : list of tuples.
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        x_train : array-like (n, dim)
            n points with dim dimensions

        Returns
        -------
        array-like (1, dim):
            Returns a 2D array. dim is the dimension of a single point
            row corresponds to a single point.
            column corresponds to a dimension.

        """
        return self.sampling_method.sample(bounds=bounds, n=1, x_history=x_train)

    @abstractmethod
    def get_point(self, x_train, y_train, iteration, bounds):
        return NotImplementedError



import scipy.linalg
import numpy as np


def inv_stable(A):
    """Inverts a quadratic matrix with non-negative eigenvalues
    Inverts using cholesky decomposition if matrix is positive semi-definite.
    If not, inverts matrix using eigenvalue decomposition.
    If eigendecomposition does not succeed then use numpy.linalg.inv

    Parameters
    ----------
    A : array-like (n, n)
        Matrix A has non-negative eigenvalues

    Returns
    -------
    array-like (n, n)
    """
    try:
        return inv_chol(A, symmetric=True)
    except:
        try:
            return inv_eig(A)
        except:
            return np.linalg.inv(A)


def inv_eig(A):
    """Inverts a quadratic matrix with non-negative eigenvalues
    Uses eigendecomposition of a matrix.

    Parameters
    ----------
    A : array-like (n, n)

    Returns
    -------
    The inverse array-like (n, n)

    Notes
    -----
    :math:`A = Q V Q^{-1}`
    :math: `A^_{-1} = (Q V Q^{-1})^{-1} = Q^{-1} V^{-1} Q`
    """
    try:
        V, Q = np.linalg.eig(A)
        V = np.diag(1/V)

        # test symmetry
        symmetric = np.allclose(A, A.T, rtol=1e-3, atol=1e-4)

        if symmetric:
            return Q @ V @ Q.T
        else:
            return Q @ V @ np.linalg.inv(Q)
    except:
        raise Exception('Finding inverse of matrix with eigendecomposition failed.')


def inv_chol(A, symmetric=None):
    """Inverts a positive semi-definite matrix A (stable)
    Adds a small positive definite matrix and uses cholesky decomposition.

    Parameters
    ----------
    A : positive semi-definite matrix (n, n)
    symmetric : boolean
        set True if matrix is symmetric. Else false.

    Returns
    -------
    An approximate inverse positive definite matrix (n, n)
    """
    not_2d = len(A.shape) != 2
    try:
        not_quadratic = A.shape[0] != A.shape[1]
    except:
        not_quadratic = True

    # test symmetry
    if symmetric is None:
        symmetric = np.allclose(A, A.T, rtol=1e-3, atol=1e-4)

    if not_2d or not_quadratic:
        raise Exception('Given Matrix is either not 2D or quadratic.')
    elif not symmetric:
        raise Exception('Given matrix is not symmetric.')
    else:
        # make A positive definite for the triangulation
        Id = np.eye(A.shape[0])
        P = Id * 1e-7
        A_tilde = A + P
        # triangulate matrix with cholesky
        L = scipy.linalg.cholesky(A_tilde, lower=True)
        L_inv = scipy.linalg.solve_triangular(L, Id, lower=True)
        A_tilde_inv = L_inv.T @ L_inv

        return A_tilde_inv



# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Datatype


class Ordinal(Datatype):
    """Ordinal Datatype
    Can be used for ordinal categories. A list of strings is used as input. The order is given by the list.
    An example to define age categories:

    >>> from grAdapt.space.datatype import Ordinal
    >>> age_category = Ordinal(['young', 'middleaged', 'old'])
    >>> bounds = [age_category]

    Internally, each category is treated as an Integer.
    """

    def __init__(self, list_str, prior='uniform'):
        """

        Parameters
        ----------
        list_str : list of strings
            Each string represents a category
        prior: str
            - if 'uniform' then nothing is changed
            - if 'log-uniform' then samples between low and high are drawn from the log-uniform distribution
        """

        self.len = len(list_str)
        self.prior = prior
        self.bound = [0, self.len-1]
        self.low = 1
        self.high = self.len
        self.categories = list_str
        self.dtype = 'ordinal'

    def __len__(self):
        return self.len

    def __getitem__(self, key):
        return self.bound[key]

    def __setitem__(self, key, value):
        self.bound[key] = value

    def get_value(self, value):
        return self.get_category(value)

    def get_category(self, value):
        key = int(np.round(value))
        return self.categories[key]

    def __repr__(self):
        return 'Ordinal('+str(self.categories)+')'

    def __str__(self):
        return 'Ordinal('+str(self.categories)+')'

    def transform(self, x):
        """

        Parameters
        ----------
        x : numeric

        Returns
        -------
        numeric

        """
        if self.prior == 'log-uniform':
            # Inverse transform sampling using quantile function
            x_normalized = x / self.high
            return self.get_value(self.low * np.power(self.high * 1.0 / self.low, x_normalized))
        else:
            return self.get_value(x)



# Python Standard Libraries
from abc import abstractmethod

# Third party imports
import numpy as np

from .base import Optimizer


class Momentum(Optimizer):
    """Optimizer object

    Implementation of Momentum (Polyak, 1964)

    """
    def __init__(self, surrogate, params=[1e-3, 0.9]):
        super().__init__(surrogate, params)

    def run(self, xp, num_iters, surrogate_grad_params):
        alpha = self.params[0]
        beta = self.params[1]
        x_next = xp
        mt = 0

        for i in range(num_iters):
            # grad has shape (d, )
            grad = self.surrogate.eval_gradient(x_next, surrogate_grad_params)
            x_old = x_next
            mt = beta * mt + (1 - beta) * grad
            x_next = x_next - alpha * mt

            # convergence
            if np.linalg.norm(x_old - x_next) < 1e-3:
                return x_next
        return x_next


class MomentumBisection(Optimizer):
    """Optimizer object

    Implementation of Momentum (Polyak, 1964)
    Modified with bisection method if a sign change of gradient happened

    """

    def __init__(self, surrogate, params=[1e-3, 0.9]):
        super().__init__(surrogate, params)

    def run(self, xp, num_iters, surrogate_grad_params, k=2):
        alpha = np.ones_like(xp) * self.params[0]
        beta = self.params[1]
        mt = 0
        x_next = xp
        grad = self.surrogate.eval_gradient(xp, surrogate_grad_params)

        for i in range(1, num_iters):
            # store old point with its gradient
            x_old = x_next
            grad_old = grad
            # Momentum update rule
            mt = beta * mt + (1 - beta) * grad_old
            x_new = x_next - alpha * mt
            grad = self.surrogate.eval_gradient(x_new, surrogate_grad_params)

            # Bisection
            # sign changed happened if sign is negative
            sign = grad * grad_old
            sign_changed = sign < 0
            # adapt alpha learning rate
            alpha = sign_changed * alpha / (k**2) + (1 - sign_changed) * alpha * k
            # x_next lies between x_old and x_new
            x_next = (1 - sign_changed) * x_new + sign_changed * (x_new + x_old) / 2
            # convergence
            if np.linalg.norm(x_old - x_next) < 1e-3:
                return x_next
        return x_next



# decorators
from abc import abstractmethod


class Datatype:
    
    @abstractmethod
    def __init__(self, *args):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass

    @abstractmethod
    def get_value(self, value):
        pass

    @abstractmethod
    def transform(self, x):
        pass

    @abstractmethod
    def inverse_transform(self, x):
        pass


# Python Standard Libraries
from abc import abstractmethod

# Third party imports
import numpy as np


class Optimizer:
    """Gradient-based Optimization Method Base class

    Parameters
    ----------
    surrogate : surrogate Object
        Surrogate model i.e. Gaussian Process Regression, Random Forrest Regression
    params : list or tuple
        List or tuple of parameters specific for the Optimizer.

    Attributes
    ----------
    self.surrogate : surrogate
    self.params : params


    Examples
    --------
    >>> optimizer = Optimizer(surrogate, params)
    >>> optimizer.run(np.array([1, 2, 3.4]), 100, [x_train])
    """

    def __init__(self, surrogate, params):
        self.surrogate = surrogate
        self.params = params

    def __call__(self, surrogate, params=None):
        self.surrogate = surrogate

        if params is not None:
            self.params = params

    def set_params(self, params):
        self.params = params

    @abstractmethod
    def run(self, xp, num_iters, surrogate_grad_params):
        """Run the optimizer from the starting point xp with n_iters steps

        Parameters
        ----------
        xp : 1D array-like (d,)
            Starting point for the Optimizer.
        num_iters : int
            Number of steps to be made.
        surrogate_grad_params : list or tuple
            List or tuple of parameters for the surrogate model. Different surrogate models require different parameters
            to evaluate the gradient.

        Returns
        -------
        x_next : 1D array-like

        """
        raise NotImplementedError


from grAdapt.models import Sequential
import numpy as np
from utils import timer, enablePrint
from grAdapt.surrogate import GPR, GPRSlidingWindow, GPROnlineInsert
from grAdapt.surrogate.kernels import RationalQuadratic, RBF, Nystroem
#import warnings
#warnings.filterwarnings("error")

def rastrigin(x):
    x = np.array(x)
    return 10*len(x)+np.sum(x**2-10*np.cos(2*np.pi*x), axis=0)
    
def sphereMin(x):
    return np.sum((x)**2)

@timer
def test():
    # Kernel Tests
    # Stock Settings
    gpr = GPR()
    model = Sequential(surrogate=gpr, random_state=1)
    bounds = [(-5, 5) for i in range(2)]
    res = model.minimize(rastrigin, bounds, 50, show_progressbar=False,
                        n_random_starts = 3)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    
    # RationalQuadratic Test
    kernel = RationalQuadratic()
    gpr = GPR(kernel=kernel)
    model = Sequential(surrogate=gpr, random_state=1)
    bounds = [(-5, 5) for i in range(2)]
    res = model.minimize(rastrigin, bounds, 50, show_progressbar=False,
                        n_random_starts = 3)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    
    # RBF Test
    gpr = GPR(kernel=RBF())
    model = Sequential(surrogate=gpr, random_state=1)
    bounds = [(-5, 5) for i in range(2)]
    res = model.minimize(rastrigin, bounds, 50, show_progressbar=False,
                        n_random_starts = 3)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    
    # Nystroem Test
    gpr = GPROnlineInsert(kernel=Nystroem(RBF()))
    model = Sequential(surrogate=gpr, random_state=1)
    bounds = [(-5, 5) for i in range(2)]
    res = model.minimize(rastrigin, bounds, 50, show_progressbar=False,
                        n_random_starts = 3)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    
    # Surrogate Tests
    # GPROnlineInsert
    gpr = GPROnlineInsert(kernel=RationalQuadratic())
    model = Sequential(surrogate=gpr, random_state=1)
    bounds = [(-5, 5) for i in range(2)]
    res = model.minimize(rastrigin, bounds, 100, show_progressbar=False,
                        n_random_starts = 3)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    
    # GPRSlidingWindow
    gpr = GPRSlidingWindow(kernel=RationalQuadratic(), window_size=25)
    model = Sequential(surrogate=gpr, random_state=1)
    bounds = [(-5, 5) for i in range(2)]
    res = model.minimize(rastrigin, bounds, 100, show_progressbar=False,
                        n_random_starts = 3)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
   
    

    
def main():
    try:
        test()
        print('Changing kernels \t \t Ok.')
    except:
        enablePrint()
        print('Changing kernels \t \t Not Ok.')
main()


# Python Standard Libraries
import numpy as np

# grAdapt
from .base import Equidistributed
from grAdapt.utils.sampling import sample_points_bounds, bounds_range
from grAdapt.utils.math.spatial import pairwise_distances


class Density(Equidistributed):
    """Density-based sampling

    For each n_candidates candidate, the number of points withing eps is determined.
    The candidate with the lowest number of points within the eps-radius is sampled.
    Repeated for n.

    Eps is determined by 1/25 * (largest bound).
    Empirically satisfying results with such setting.

    Disadvantages: Creates holes, not reliable in the beginning.
    """

    def __init__(self, n_candidates=5, eps='auto', window_size=500):
        """
        Parameters
        ----------
        n_candidates : integer
            number of candidates
        eps : int or string
            if set to 'auto', then 1/25 of the largest bound is set as eps
            else eps is set as given
        window_size : integer
            size of history points to consider. smaller is faster but worsen the results.
        """
        super().__init__()
        self.n_candidates = n_candidates
        self.eps = eps
        self.window_size = window_size

    def sample(self, bounds, n, x_history=None):
        """Samples low discrepancy/equidistributed sequences
        Method has to handle with new bounds and n.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n : int
            number of points to be sampled
        x_history : array-like (2d)
            History points. Consider those to prevent sampling in dense regions.

        Returns
        -------
        array-like (n, len(bounds))
            Returns a 2D array. dim is the dimension of a single point
            Each row corresponds to a single point.
            Each column corresponds to a dimension.
        """

        # set to new variables
        super().sample(bounds, n, x_history)

        if x_history is None:
            x_history_list = list(sample_points_bounds(self.bounds, 1))
        else:
            x_history_list = list(x_history)

        if self.eps == 'auto':
            self.eps = 1 / 25 * np.max(np.array(list(map(bounds_range, self.bounds))))

        start_eps = self.eps

        for i in range(self.n):

            # auto decrease
            self.eps = start_eps/np.log(np.e+i)

            candidates = sample_points_bounds(self.bounds, self.n_candidates)
            dists_matrix = pairwise_distances(candidates, np.array(x_history_list[-self.window_size:]))
            clusters = dists_matrix < self.eps  # cluster for every point in p
            clusters_size = np.sum(clusters, axis=1)  # calc size for each cluster
            # first point tends to be empty already (sparsity at the beginning)
            if clusters_size[0] == 0:
                x_history_list.append(candidates[0])
            else:
                smallest_cluster_idx = np.argmin(clusters_size)
                x_history_list.append(candidates[smallest_cluster_idx])  # add to current points

        return np.array(x_history_list)[-self.n:]



# Python Standard Libraries
import numpy as np

# decorators
from abc import abstractmethod

# Third party imports
from scipy.stats import loguniform

# grAdapt


class Transformer:

    def __init__(self, func, bounds, transform='identity'):
        """

        Parameters
        ----------
        func : function to transform the input
        bounds : list of Datatype's
        transform : str
            `identity` : no changes are applied
            `normalize` : normalize each dimension of bounds to [0, 1]
        """
        self.func = func
        self.bounds = bounds

    def __call__(self, x):
        """

        Parameters
        ----------
        x : array-like (d, )

        Returns
        -------

        """
        x_transformed = np.zeros_like(x, dtype='O')

        for i in range(x.shape[0]):
            try:
                x_transformed[i] = self.bounds[i].transform(x[i])
            except:
                x_transformed[i] = x[i]

        if all(isinstance(x, float) for x in x_transformed):
            x_transformed = np.array(x_transformed, dtype=np.float)

        return self.func(x_transformed)



# Python Standard Libraries
import numpy as np

# grAdapt package
from .base import Escape, inside_bounds, sample_points_bounds, bounds_range_ndim


class NormalDistribution(Escape):
    """Standard Multivariate Normal Distribution
    is added to the current best position

    """

    def __init__(self, surrogate, sampling_method=None):
        super().__init__(surrogate, sampling_method)

    def get_point(self, x_train, y_train, iteration, bounds):
        """
        Parameters
        ----------
        self : self object
        x_train : array-like shape (n, d)
        y_train : array-like shape (n,)
        iteration : integer
        bounds: list of 2-tuples

        Returns
        -------
        array-like (d,)
        """
        x_best = x_train[np.argmin(y_train)]
        cov_matrix = np.diag((bounds_range_ndim(bounds)/6)**2)

        for i in range(20):
            x_next = np.random.multivariate_normal(x_best, cov_matrix)
            if inside_bounds(bounds, x_next):
                return x_next

        return self.escape_history(bounds, x_train)



from grAdapt.models import Sequential
import numpy as np
from utils import timer, enablePrint
from grAdapt.surrogate import GPR
from grAdapt.surrogate.kernels import RationalQuadratic 

def rastrigin(x):
    x = np.array(x)
    return 10*len(x)+np.sum(x**2-10*np.cos(2*np.pi*x), axis=0)
    
def sphereMin(x):
    return np.sum(x**2)

@timer
def test():
    #gpr = GPR(kernel=RationalQuadratic())
    #model = Sequential(surrogate=gpr)
    model = Sequential()
    bounds = [(-10, 10) for i in range(5)]
    res = model.minimize(rastrigin, bounds, 10, show_progressbar=False,
                         auto_checkpoint=True)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    print('Argmin: {xmin}'.format(xmin=x[np.argmin(y)]))
    res2 = model.load_checkpoint(model.checkpoint_file)
    
    assert (res2['x'] == res['x']).all()
    assert (res2['y'] == res['y']).all()
    assert (res2['x_sol'] == res['x_sol']).all()
    
def main():
    try:
        test()
        print('Loading Checkpoint \t \t Ok.')
    except:
        enablePrint()
        print('Loading Checkpoint \t \t Not Ok.')
main()


# Python Standard Libraries
# decorators
from deprecated import deprecated

# Third party imports
import numpy as np

# Package imports
from .base import Surrogate
from .kernels import RBF


class RBFRegression(Surrogate):

    @deprecated("Deprecated object name. The name of this object might be misleading.")
    def __init__(self, **surrogate_params):
        super().__init__(None, surrogate_params)
        self.RBF = RBF()
        self.X_train = None
        self.alpha_ = None

    def fit(self, X, y):
        """Fit training data (X, y)
        Parameters
        ----------
        X : array-like (n, d)
        y : array-like (n,)


        Returns
        -------
        None

        """
        self.alpha_ = y

        if len(X.shape) == 1:
            self.X_train = X.reshape(1, -1)
        else:
            self.X_train = X

    def predict(self, x_predict, return_std=False):
        """Predict y value based on x_predict
        Parameters
        ----------
        x_predict : array-like (n, d)
        return_std : array-like (n,)
            returns the standard deviation of the posterior mean


        Returns
        -------
        array-like (n,)

        """
        if len(x_predict.shape) == 1:
            x_predict = x_predict.reshape(1, -1)
        return self.RBF(x_predict, self.X_train) @ self.alpha_

    def eval_gradient(self, x, surrogate_grad_params):
        """Evaluates the gradient of GPR at point x
        Parameters
        ----------
        x: 1D array-like (d,)
        surrogate_grad_params: list
            surrogate_grad_params[0]: whole training data shape (m, d)


        Returns
        -------
        grad : 1D array-like
            same shape as x
        """

        x_train = surrogate_grad_params[0]
        x = x.reshape(1, -1)
        kernel_matrix = self.RBF(x, x_train)
        length_scale = self.RBF.length_scale
        n, m = kernel_matrix.shape[0], kernel_matrix.shape[1]
        d = x.shape[1]

        grad = np.zeros(d)
        # calculate gradient
        for k in range(d):
            tmp_kernel = np.copy(kernel_matrix)
            tmp_kernel[0, :] = ((x_train[:, k] - x[0, k]) * tmp_kernel[0, :]) / (length_scale ** 2)

            grad[k] = tmp_kernel.dot(self.alpha_)

        return grad

    def kernel_(self, x_test, x_train):
        """Call the kernel function
        Parameters
        ----------
        x_test : array-like (n, d)
        x_train : array-like (m, d)


        Returns
        -------
        array-like (n, m)
        """

        if len(x_test.shape) == 1:
            x_test = x_test.reshape(1, -1)

        if len(x_train.shape) == 1:
            x_train = x_train.reshape(1, -1)

        return self.RBF(x_test, x_train)



import unittest
from grAdapt.sampling.equidistributed import Density, KDistance, MaximalMinDistance, Mitchell, Reassignment
from grAdapt.utils.sampling import inside_bounds_ndim

import numpy as np


class TestEquidistributed(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        self.sampling_methods = [Density(), KDistance(), MaximalMinDistance(), Mitchell(), Reassignment()]
        super().__init__(*args, **kwargs)

    def test_sample_number(self):
        """
        Test all sampling methods in self.sampling_methods.
        Tests for several dimension and sample sizes
        whether the first-axis of the output equals
        the sample size n.
        """
        n_samples = 33
        dims = np.random.randint(1, 40, 10)

        for sampling in self.sampling_methods:
            for i in range(1, n_samples):
                for d in dims:
                    bounds = [(-10, 10) for i in range(d)]
                    sample = sampling.sample(bounds, i)
                    self.assertEqual(sample.shape[0], i)


if __name__ == '__main__':
    unittest.main()



# python
# import warnings

# Third party imports
import numpy as np

# grAdapt
from .base import Initial
from grAdapt.utils.sampling import sample_corner_bounds


class Vertices(Initial):
    """
    Samples vertices if n_evals >= 2 ** len(bounds).
    Else low discrepancy sequences are sampled.
    """

    def __init__(self, sampling_method):
        """
        Parameters
        ----------
        sampling_method : grAdapt.sampling.equidistributed Object
            Sample low discrepancy sequences when initial point method is not feasible
        """
        super().__init__(sampling_method)

    def sample(self, bounds, n_evals):
        """Returns a numpy array of sampled points.
        Does not include corner points of the hypercube/search space.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n_evals : int
            number of initial points sampled by method

        Returns
        -------
        (self.n_evals, len(self.bounds)) numpy array
        """
        super().sample(bounds, n_evals)
        if 2 ** len(self.bounds) > self.n_evals:
            return self.sampling_method.sample(bounds=bounds, n=n_evals)
        else:
            corner_points = sample_corner_bounds(self.bounds)
            num_corner_points = corner_points.shape[0]
            if self.n_evals > 2 ** len(self.bounds):
                random_points = self.sampling_method.sample(bounds=self.bounds,
                                                            n=(self.n_evals - num_corner_points),
                                                            x_history=corner_points)
                return np.vstack((corner_points, random_points))
            else:
                return corner_points



# Python Standard Libraries
import numpy as np

# grAdapt package
from .base import Escape, inside_bounds, sample_points_bounds, bounds_range_ndim

# TODO
class CMAES(Escape):

    def __init__(self, surrogate, sampling_method=None):
        super().__init__(surrogate, sampling_method)

    def get_point(self, x_train, y_train, iteration, bounds):
        raise NotImplementedError('CMAES has not been implemented yet.')



import os
import utils

# add new test scripts here
files = ['basic.py', 'changing_kernels.py', 'load_checkpoint.py', 'maximize.py', 'slidingWindow.py', 'training_continuation.py']

# filter '.py' files endings
filtered_files = []
for file in files:
    exclude_files = ['__init__.py', 'run_all.py', 'test.py']
    exclude_file_extension = ['.pyc']
    
    if file in exclude_files:
        continue
    
    if any(list(map(lambda x: x in file, exclude_file_extension))):
        continue
        
    if file[-3:] != '.py':
        continue
        
    filtered_files.append(file)

print(filtered_files)
print(len(filtered_files))

# import
counter = 0
for file in filtered_files:
    mod = __import__(file[:-3], locals(), globals())
    utils.enablePrint()
    # out = mod.main()
    # print(out)


import unittest
from grAdapt.models import Sequential

import numpy as np


class TestTrainingContinuationEqualValues(unittest.TestCase):

    @staticmethod
    def rastrigin(x):
        x = np.array(x)
        return 10 * len(x) + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x), axis=0)

    @staticmethod
    def sphere(x):
        return np.sum(x ** 2)

    @staticmethod
    def equal_function(function):
        model = Sequential()
        bounds = [(-10, 10) for _ in range(5)]
        n1 = 30
        n2 = 50
        model.minimize(function, bounds, n1, show_progressbar=False)
        res = model.minimize(function, bounds, n2, show_progressbar=False)

        x_train = res['x']
        y_train = res['y']

        y_compare = np.array(list(map(function, x_train)))
        
        print('x_train:')
        print(x_train)
        print('y_sol')
        print(res['y_sol'])

        return y_train, y_compare, x_train

    def test_equal_sphere(self):
        y_train, y_compare, x_train = self.equal_function(function=self.sphere)
        self.assertTrue(np.array_equal(y_compare, y_train))

    def test_equal_rastrigin(self):
        y_train, y_compare, x_train = self.equal_function(function=self.rastrigin)
        self.assertTrue(np.array_equal(y_compare, y_train))


if __name__ == '__main__':
    unittest.main()



from .base import Escape

# Import Subclasses
from .MatyasNormalDistribution import MatyasNormalDistribution
from .MaxStandardDeviation import MaxStandardDeviation
from .NormalDistribution import NormalDistribution
from .NormalDistributionDecay import NormalDistributionDecay
from .StandardNormalDistribution import StandardNormalDistribution
from .UniformDistribution import UniformDistribution



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(name='grAdapt',
      version='0.1.1b4',
      description='grAdapt: Gradient Adaption for Black-Box Optimization.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Manh Khoi Duong',
      author_email='manh.duong@hhu.de',
      url='https://github.com/mkduong-ai/grAdapt',
      download_url = 'https://github.com/mkduong-ai/grAdapt/blob/master/dist/grAdapt-0.1.1b4.tar.gz?raw=true',
      keywords = ['grAdapt','black-box optimization', 'optimization', 'smbo', 'hyperparameter optimization', 'hyperparameter', 'sequential model-based optimization', 'stochastic optimization', 'global optimization', 'machine learning', 'toolbox'],
      license='Apache License 2.0',
      # py_modules=['grAdapt'],
      packages=find_packages(),
      # package_dir={'': 'src'},
      classifiers=[
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
      ],
      install_requires=[
          'numpy ~= 1.18',
          'scipy ~= 1.4',
          'scikit-learn ~= 0.22',
          'tqdm ~= 4.44',
          'deprecated ~= 1.2.7',
      ],
      )


import unittest
from grAdapt.models import Sequential

import numpy as np


class TestTrainingContinuationEqualValues(unittest.TestCase):

    @staticmethod
    def rastrigin(x):
        x = np.array(x)
        return 10 * len(x) + np.sum(x ** 2 - 10 * np.cos(2 * np.pi * x), axis=0)

    @staticmethod
    def sphere(x):
        return np.sum(x ** 2)

    @staticmethod
    def equal_function(function):
        bounds = [(-10, 10) for _ in range(5)]
        n1 = 100
        
        # first experiment
        model = Sequential(random_state=1234)
        model.minimize(function, bounds, n1, show_progressbar=False)
        res = model.res
        
        x_train1 = res['x']
        y_train1 = res['y']

        # second experiment
        model = Sequential(random_state=1234)
        model.minimize(function, bounds, n1, show_progressbar=False)
        res = model.res
        
        x_train2 = res['x']
        y_train2 = res['y']
        
        
        return x_train1, x_train2, y_train1, y_train2

    def test_equal_sphere(self):
        x_train1, x_train2, y_train1, y_train2 = self.equal_function(function=self.sphere)
        self.assertTrue(np.array_equal(x_train1, x_train2))
        self.assertTrue(np.array_equal(y_train1, y_train2))
        
    def test_equal_rastrigin(self):
        x_train1, x_train2, y_train1, y_train2 = self.equal_function(function=self.rastrigin)
        self.assertTrue(np.array_equal(x_train1, x_train2))
        self.assertTrue(np.array_equal(y_train1, y_train2))
        

if __name__ == '__main__':
    unittest.main()



from grAdapt.models import Sequential
import numpy as np
from utils import timer, enablePrint

def rastrigin(x):
    x = np.array(x)
    return 10*len(x)+np.sum(x**2-10*np.cos(2*np.pi*x), axis=0)

def sphere(x):
    return np.sum(x**2)

@timer
def test():
    model = Sequential()
    bounds = [(-10, 10) for i in range(10)]
    n = 25
    res = model.minimize(sphere, bounds, n, show_progressbar=False)
    x = res['x']
    y = res['y']
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    # print('Argmin: {xmin}'.format(xmin=x[np.argmin(y)]))
    print('')
    print('Training continuation...')
    training = (x, y)
    res = model.minimize(sphere, bounds, n, show_progressbar=False)
    x = res['x']
    y = res['y']
    y_real = np.array(list(map(sphere, x)))
    print('Minimum found: {ymin}'.format(ymin=np.min(y)))
    
    if np.max(np.abs(y_real-y)) >= 1e-10:
        print('Max single error: {}'.formatnp.max(np.abs(y_real-y)))
        raise Exception('y Values wrong and not equal!')
    
    if len(x) != 2*n:
        print(len(x))
        raise Exception('Training continuation FAILED.')
    
def main():
    try:
        test()
        print('Training continuation \t \t Ok.')
    except:
        enablePrint()
        print('Training continuation \t \t Not Ok.')
main()


from .base import Surrogate

from .GPR import GPR, GPRSlidingWindow, GPROnlineInsert, GPROnlineStandard
from .NoGradient import NoGradient
from .NoModel import NoModel
from .RBFRegression import RBFRegression



# decorators
from abc import abstractmethod


class Equidistributed:
    """
    Equidistributed Base class
    It samples points with sample() by given bounds and n
    """

    def __init__(self):
        """
        Parameters
        ----------

        Notes
        -----
        All subclasses can be initialized without parsing arguments.
        """
        self.bounds = None
        self.n = None
        self.window_size = 0

    @abstractmethod
    def sample(self, bounds, n, x_history=None):
        """Samples low discrepancy/equidistributed sequences
        Method has to handle with new bounds and n_evals.

        Parameters
        ----------
        bounds : list of tuples or list of grAdapt.space.datatype.base
            Each tuple in the list defines the bounds for the corresponding variable
            Example: [(1, 2), (2, 3), (-1, 4)...]
        n : int
            number of points to be sampled
        x_history : 2d array-like
            History points. Consider those to prevent sampling in dense regions.

        Returns
        -------
        (n, len(bounds)) numpy array
        """
        self.bounds = bounds
        self.n = n

        # raise Exceptions
        if self.n == 0:
            raise ValueError('n is zero. You must sample with a positive n.')



# Python Standard Libraries
from abc import abstractmethod

# Third party imports
import numpy as np

from .base import Optimizer


class Adam(Optimizer):
    """Optimizer object

    Implementation of ADAM (Kingma et al.)

    """

    def __init__(self, surrogate, params=[1e-3, 0.9, 0.999, 1e-8]):
        super().__init__(surrogate, params)

    def run(self, xp, num_iters, surrogate_grad_params):
        alpha = self.params[0]
        beta_1 = self.params[1]
        beta_2 = self.params[2]
        epsilon = self.params[3]
        m_t = 0
        v_t = 0
        x_next = xp

        for t in range(1, num_iters + 1):
            g_t = self.surrogate.eval_gradient(x_next, surrogate_grad_params)
            x_old = x_next

            # adam update rule
            m_t = beta_1 * m_t + (1 - beta_1) * g_t
            v_t = beta_2 * v_t + (1 - beta_2) * (g_t * g_t)
            m_cap = m_t / (1 - (beta_1 ** t))
            v_cap = v_t / (1 - (beta_2 ** t))
            x_next = x_next - (alpha * m_cap) / (np.sqrt(v_cap) + epsilon)  # update step

            if np.linalg.norm(x_next - x_old) < 1e-3:
                return x_next
        return x_next


class AdamBisection(Optimizer):
    """Optimizer object

    Implementation of ADAM (Kingma et al.)
    Modified with bisection method if a sign change of gradient happened

    """

    def __init__(self, surrogate, params=[1e-3, 0.9, 0.999, 1e-8]):
        super().__init__(surrogate, params)

    def run(self, xp, num_iters, surrogate_grad_params, k=1):
        alpha = np.ones_like(xp) * self.params[0]
        beta_1 = self.params[1]
        beta_2 = self.params[2]
        epsilon = self.params[3]
        m_t = 0
        v_t = 0
        x_next = xp
        g_t = self.surrogate.eval_gradient(x_next, surrogate_grad_params)

        for t in range(2, num_iters + 1):
            # store old point with its gradient
            x_old = x_next
            g_t_old = g_t

            # ADAM update rule
            m_t = beta_1 * m_t + (1 - beta_1) * g_t_old
            v_t = beta_2 * v_t + (1 - beta_2) * (g_t_old * g_t_old)
            m_cap = m_t / (1 - (beta_1 ** t))
            v_cap = v_t / (1 - (beta_2 ** t))
            x_new = x_next - (alpha * m_cap) / (np.sqrt(v_cap) + epsilon)  # update step

            # gradient update
            g_t = self.surrogate.eval_gradient(x_new, surrogate_grad_params)

            # Bisection
            # sign changed happened if sign is negative
            sign = g_t * g_t_old
            sign_changed = sign < 0
            # adapt alpha learning rate
            alpha = sign_changed * alpha / (k**2) + (1 - sign_changed) * alpha * k
            # x_next lies between x_old and x_new
            x_next = (1 - sign_changed) * x_new + sign_changed * (x_new + x_old) / 2
            # convergence
            if np.linalg.norm(x_next - x_old) < 1e-3:
                return x_next
        return x_next



import unittest
from grAdapt.sampling.equidistributed import Density, KDistance, MaximalMinDistance, Mitchell, Reassignment
from grAdapt.utils.sampling import inside_bounds_ndim

import numpy as np


class TestEquidistributed(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        self.sampling_methods = [Density(), KDistance(), MaximalMinDistance(), Mitchell(), Reassignment()]
        super().__init__(*args, **kwargs)

    @staticmethod
    def random_range(low, high, size):
        return np.random.rand(size)*(high-low)+low

    def test_inside_bounds(self):
        """
        Test all sampling methods in self.sampling_methods.
        Tests for several dimension and sample sizes
        whether each row is inside the given bound.
        """

        n_samples = np.random.randint(1, 128, 32)
        dims = np.random.randint(1, 20, 8)

        for sampling_method in self.sampling_methods:
            for i in n_samples:
                for d in dims:
                    lows = self.random_range(-10000, 0, d)
                    highs = self.random_range(0, 10000, d)
                    bounds = [(low, high) for low, high in zip(lows, highs)]
                    sample = sampling_method.sample(bounds, i)
                    self.assertEqual(True, inside_bounds_ndim(bounds, sample))
                    del sample


if __name__ == '__main__':
    unittest.main()
